-- Pulls back the Addon-Local Variables and store them locally.
local addonName, addonTable = ...;
local addon = _G[addonName];
local L = LibStub("AceLocale-3.0"):GetLocale(addonName);

-- Store local copy of Blizzard API wrapper and other addon global functions and variables
local Blizzard, systemEvents, systemTimers = addon.API.Blizzard, addon.System.Events, addon.System.Timers;
local ipairs, pairs, tinsert, unpack = ipairs, pairs, tinsert, unpack;

local trackedUnits = {
	pet = true,
	player = true,
	target = true,
};

local AuraCache, persistentMultipliers, playerAuras, prevGUID, TotemTimes = {}, {}, {}, {}, {};
local Auras, Events = {}, {};

local function calculatePandemic(spellID)
	return addon.Core.Tooltip.GetSpellBaseDuration(spellID) * 0.3;
end

local function processTotems(spellID)
	local coreTables = addon.Core.Tables;
	local auraTable = coreTables.GetValue(addonTable.TotemData, spellID);

	if not coreTables.IsEmpty(auraTable) then
		local spellInfo = addon.API.Spell.GetSpellInfo(spellID);
		local endTime = Blizzard.GetTime() + coreTables.GetValue(spellInfo, "CastingTime", 0) + addon.Core.Tooltip.GetSpellBaseDuration(spellID);

		if not coreTables.IsEmpty(auraTable) then
			for _, auraID in ipairs(auraTable) do
				TotemTimes[auraID] = endTime;
			end
		else
			TotemTimes[auraTable] = endTime;
		end
	end
end

Auras = {
	Add = function(unitID, GUID, auraTable, section, index,  ...)
		local coreTables = addon.Core.Tables;

		if GUID and auraTable and section then
			local auraName, auraIcon, auraStacks, auraType, auraDuration, expirationTime, auraCaster, isStealable, _, spellID, canApplyAura, isBossAura, _, _, timeMod = unpack(...);

			if spellID then
				local returnValue = {
					ID = spellID,
					Name = auraName,
					IconID = auraIcon,
					Stacks = auraStacks,
					Type = auraType,
					Start = (expirationTime or 0) - (auraDuration or 0),
					Duration = auraDuration or 0,
					End = expirationTime or 0,
					PandemicTime = calculatePandemic(spellID),
					Caster = auraCaster,
					IsStealable = isStealable,
					CanApplyAura = canApplyAura,
					IsBossAura = isBossAura,
					TimeMod = timeMod,
				};

				if auraCaster == "player" or "pet" then
					if trackedUnits[unitID] then
						playerAuras[GUID] = playerAuras[GUID] or {};
						playerAuras[GUID][spellID] = true;
					end

					if coreTables.GetValue(persistentMultipliers[GUID], auraName) then
						returnValue.PersistentMultiplier = persistentMultipliers[GUID][auraName];
					end

					auraTable.Player[section][index] = returnValue;

					auraTable.Player[section].IDS[spellID] = auraTable.Player[section].IDS[spellID] or {};
					auraTable.Player[section].NAMES[auraName] = auraTable.Player[section].NAMES[auraName] or {};

					tinsert(auraTable.Player[section].IDS[spellID], index);
					tinsert(auraTable.Player[section].NAMES[auraName], index);
				else
					auraTable.Other[section][index] = returnValue;

					auraTable.Other[section].IDS[spellID] = auraTable.Other[section].IDS[spellID] or {};
					auraTable.Other[section].NAMES[auraName] = auraTable.Other[section].NAMES[auraName] or {};

					tinsert(auraTable.Other[section].IDS[spellID], index);
					tinsert(auraTable.Other[section].NAMES[auraName], index);
				end

				return true;
			end
		end

		return false;
	end,

	AddBuff = function(unitID, GUID, index)
		if GUID then
			return Auras.Add(unitID, GUID, AuraCache[GUID], "Buff", index, { Blizzard.UnitBuff(unitID, index) });
		end

		return false;
	end,

	AddDebuff = function(unitID, GUID, index)
		if GUID then
			return Auras.Add(unitID, GUID, AuraCache[GUID], "Debuff", index, { Blizzard.UnitDebuff(unitID, index) });
		end

		return false;
	end,

	Get = function(auraTable, spellID, exactMatch)
		local coreTables = addon.Core.Tables;

		if not coreTables.IsEmpty(auraTable) and spellID then
			local tableIDs = coreTables.GetValue(auraTable.IDS, spellID);

			if not coreTables.IsEmpty(tableIDs) then
				return coreTables.GetValue(auraTable, tableIDs[1]);
			end

			if not exactMatch then
				local spellName = Blizzard.GetSpellInfo(spellID);

				tableIDs = coreTables.GetValue(auraTable.NAMES, spellName);

				if not coreTables.IsEmpty(tableIDs) then
					return coreTables.GetValue(auraTable, tableIDs[1]);
				end
			end
		end
	end,

	GetAura = function(GUID, spellID, anyCaster, exactMatch)
		if GUID and spellID then
			return Auras.GetBuff(GUID, spellID, anyCaster, exactMatch) or Auras.GetDebuff(GUID, spellID, anyCaster, exactMatch);
		end
	end,

	GetBuff = function(GUID, spellID, anyCaster, exactMatch)
		if GUID and spellID then
			return Auras.Get(AuraCache[GUID].Player.Buff, spellID, exactMatch, GUID)
			    or (anyCaster == true and Auras.Get(AuraCache[GUID].Other.Buff, spellID, exactMatch));
		end
	end,

	GetDebuff = function(GUID, spellID, anyCaster, exactMatch)
		if GUID and spellID then
			return Auras.Get(AuraCache[GUID].Player.Debuff, spellID, exactMatch)
			    or (anyCaster == true and Auras.Get(AuraCache[GUID].Other.Debuff, spellID, exactMatch));
		end
	end,

	Scan = function(unitID, GUID)
		local coreTables = addon.Core.Tables;

		if unitID and GUID then
			local previousAuras;

			if trackedUnits[unitID] then
				previousAuras = coreTables.Duplicate(coreTables.GetValue(playerAuras, GUID, {}));
				playerAuras[GUID] = {};
			end

			AuraCache[GUID] = {
				Player = { Buff = { IDS = {}, NAMES = {} },	Debuff = { IDS = {}, NAMES = {} } },
				Other = { Buff = { IDS = {}, NAMES = {} }, Debuff = { IDS = {},	NAMES = {} } },
			};

			for counter = 1, 40 do
				if not Auras.AddBuff(unitID, GUID, counter) then
					break;
				end
			end

			for counter = 1, 40 do
				if not Auras.AddDebuff(unitID, GUID, counter) then
					break;
				end
			end

			if trackedUnits[unitID] and not coreTables.Equate(previousAuras, coreTables.GetValue(playerAuras, GUID, {})) then
				addon.System.Callbacks.Callback("aurasChanged", unitID);
			end
		end
	end,
};

Events = {
	spellCast = function(...)
		local _, unitID, _, spellID = ...;

		if unitID == "player" then
			-- Process totem duration time to be used as remaining time on buffs when the buff is active.
			processTotems(spellID);
		end
	end,

	timerEvent = function()
		local coreTables = addon.Core.Tables;
		local currentTime = Blizzard.GetTime();

		if not coreTables.IsEmpty(TotemTimes) then
			for spellID, endTime in pairs(TotemTimes) do
				if endTime < currentTime then
					TotemTimes[spellID] = nil;
				end
			end
		end
	end,

	unitAura = function(...)
		local _, unitID = ...;
		local GUID = Blizzard.UnitGUID(unitID);

		if GUID and addon.Core.Tables.IsTable(AuraCache) then
			AuraCache[GUID] = nil;

			if trackedUnits[unitID] then
				Auras.Scan(unitID, GUID);
			end
		end
	end,
};

addonTable.Auras = {
	GetPersistentMultiplier = function(unitID, spellID)
		local coreTables = addon.Core.Tables;

		if unitID and spellID then
			local spellName = Blizzard.GetSpellInfo(addonTable.PersistentMultiplierData[spellID]);

			return coreTables.GetValue(coreTables.GetValue(persistentMultipliers, Blizzard.UnitGUID(unitID), {}), spellName, 1);
		end

		return 1;
	end,

	SetPersistentMultiplier = function(GUID, spellID)
		if GUID and spellID and addonTable.PersistentMultiplierData[spellID] then
			local spellName = Blizzard.GetSpellInfo(addonTable.PersistentMultiplierData[spellID]);

			persistentMultipliers[GUID] = persistentMultipliers[GUID] or {};
			persistentMultipliers[GUID][spellName] = addon.Units.Player.PersistentMultiplier(spellID);
		end
	end,

	UnitAura = function(unitID, spellID, anyCaster, exactMatch, GUID)
		GUID = GUID or Blizzard.UnitGUID(unitID);

		if GUID and spellID then
			if unitID and not AuraCache[GUID] then
				Auras.Scan(unitID, GUID);
			end

			local returnValue = Auras.GetAura(GUID, spellID, anyCaster, exactMatch);
			local totemEnd = TotemTimes[spellID];

			if returnValue and returnValue.End == 0 and totemEnd then
				returnValue.End = totemEnd;
			end

			return returnValue;
		end
	end,

	UnitBuff = function(unitID, spellID, anyCaster, exactMatch, GUID)
		GUID = GUID or Blizzard.UnitGUID(unitID);

		if GUID and spellID then
			if unitID and not AuraCache[GUID] then
				Auras.Scan(unitID, GUID);
			end

			local returnValue = Auras.GetBuff(GUID, spellID, anyCaster, exactMatch);
			local totemEnd = TotemTimes[spellID];

			if returnValue and returnValue.End == 0 and totemEnd then
				returnValue.End = totemEnd;
			end

			return returnValue;
		end
	end,

	UnitDebuff = function(unitID, spellID, anyCaster, exactMatch, GUID)
		GUID = GUID or Blizzard.UnitGUID(unitID);

		if GUID and spellID then
			if unitID and not AuraCache[GUID] then
				Auras.Scan(unitID, GUID);
			end

			local returnValue = Auras.GetDebuff(GUID, spellID, anyCaster, exactMatch);
			local totemEnd = TotemTimes[spellID];

			if returnValue and returnValue.End == 0 and totemEnd then
				returnValue.End = totemEnd;
			end

			return returnValue;
		end
	end,
};

local tableEvents = {
	UNIT_AURA = Events.unitAura,
	UNIT_SPELLCAST_START = Events.spellCast,
	UNIT_SPELLCAST_SUCCEEDED = Events.spellCast,
};

local tableTimers = {
	[1] = Events.timerEvent,
};

systemEvents.RegisterEvents(tableEvents, true);
systemTimers.RegisterTimers(tableTimers);