-- Pulls back the Addon-Local Variables and store them locally.
local addonName, addonTable = ...;
local addon = _G[addonName];
local L = LibStub("AceLocale-3.0"):GetLocale(addonName);

-- Store local copy of Blizzard API wrapper and other addon global functions and variables
local Blizzard, PowerType = addon.API.Blizzard, addonTable.Enum.PowerType;
local ceil, floor, format, huge, max, min, setmetatable, sort, tinsert = ceil, floor, format, math.huge, max, min, setmetatable, sort, tinsert;
local Cache = setmetatable({}, { __mode = "kv" });

addonTable.propertiesPower = function(Unit, powerType)
	local Objects = addon.Core.Objects;

	local unitID = Objects.GetID(Unit) or Unit;
	local key = format("%s-%s", unitID, powerType);

	local self = Cache[key] or {};

	local function castCost(spellID)
		local Player = addon.Units.Player;
		local Tables = addon.Core.Tables;
		local apiSpell = addon.API.Spell;

		local returnValue = 0;

		if spellID and unitID == "player" then
			local spellCost = apiSpell.GetSpellPowerCost(spellID);

			-- Check to make sure the power type for the cost matches our current power type
			if powerType == Tables.GetValue(spellCost, "Type", PowerType.None) then
				returnValue = Tables.GetValue(spellCost, "Cost", 0);
			end
			-- If we are casting, we check what the cost of the cast
			-- Ignore channel casts as the cost is removed at the start of the channel
		elseif unitID == "player" and Player.Casting.IsCasting() then
			local spellCost = apiSpell.GetSpellPowerCost(Player.Casting.ID());

			-- Check to make sure the power type for the cost matches our current power type
			if powerType == Tables.GetValue(spellCost, "Type", PowerType.None) then
				returnValue = Tables.GetValue(spellCost, "Cost", 0);
			end
		end

		return returnValue;
	end

	local function castGenerate(spellID)
		local Player = addon.Units.Player;
		local returnValue;

		if spellID and unitID == "player" then
			local Spell = Objects.newSpell(spellID);

			returnValue = Spell.PowerGenerated();
			-- If we are casting, we check what we will power we will receive from the cast
		elseif unitID == "player" and Player.Casting() then
			local Spell = Objects.newSpell(Player.Casting.ID());

			returnValue = Spell.PowerGenerated();
		end

		return returnValue or 0;
	end

	local function castRegen(spellID)
		local Player = addon.Units.Player;
		local Spell = Objects.newSpell(spellID);

		local returnValue = 0;

		if spellID and unitID == "player" then
			returnValue = Spell.ExecuteTime() * self.Regen();
			-- If we are casting, we check what we will regen until the end of the cast
		elseif unitID == "player" and Player.Casting() then
			returnValue = max(Player.Casting.Remains(), Player.GCD.Remains()) * self.Regen();
		end

		return returnValue;
	end

	-- Computes any rune cooldown.
	local function ComputeRuneCooldown(slot)
		local General = addon.API.General;

		-- Get rune cooldown infos
		local cooldownStart, cooldownDuration, runeReady = Blizzard.GetRuneCooldown(slot);

		if not runeReady then
			-- Compute the CD.
			local cooldown = (cooldownStart + cooldownDuration) - General.GetTime();

			-- Return the Rune CD
			return max(cooldown, 0);
		else
			-- Return 0 if the rune isn't in CD.
			return 0;
		end
	end

	-- Compute time to X runes.
	local function ComputerTimeToXRunes(numRunes)
		if self.Current() < numRunes then
			local Runes = {};

			for counter = 1, 6 do
				tinsert(Runes, ComputeRuneCooldown(counter));
			end

			sort(Runes, function(a, b) return a < b; end);

			return Runes[numRunes];
		end

		return 0;
	end

	-- Return abount of power after spell cast allowing for natural regen, spell cost and any power generated by the spell.
	-- This can return a value higher then self.Max() if you will cap power during the cast.
	self.AfterCast = function(Spell, numCasts)
		local returnValue = self.Current();

		if Spell then
			local spellID = Objects.IsSpell(Spell) and Spell.SpellID() or Spell;

			numCasts = numCasts or 1;
			returnValue = self.Current() - (castCost(spellID) * numCasts) + (castGenerate(spellID) * numCasts) + (castRegen(spellID) * numCasts);
		end

		return returnValue;
	end

	self.CastRegen = function(Spell)
		local returnValue = 0;

		if Spell then
			local spellID = Objects.IsSpell(Spell) and Spell.SpellID() or Spell;

			returnValue = castRegen(spellID);
		end

		return returnValue;
	end

	self.Current = function()
		local Tables = addon.Core.Tables;
		local apiUnit = addon.API.Unit;

		local returnValue;

		if powerType == PowerType.Runes then
			returnValue = 0;

			for counter = 1, 6 do
				if ComputeRuneCooldown(counter) == 0 then
					returnValue = returnValue + 1;
				end
			end

			return returnValue;
		else
			local powerInfo = apiUnit.GetPowerInfo(unitID, powerType);
			local currentPower = Tables.GetValue(powerInfo, "Current");

			if currentPower then
				return min(currentPower - castCost() + castGenerate() + castRegen(), self.Max());
			end

			return 1;
		end
	end

	-- power.deficit
	self.Deficit = function()
		return self.Max() - self.Current();
	end

	-- "power.deficit.pct"
	self.DeficitPercent = function()
		return (self.Deficit() / self.Max()) * 100;
	end

	if powerType == PowerType.Insanity then
		-- Insanity Drain
		self.Drain = function()
			local Insanity = addonTable.Insanity;

			-- Drain calculated by trackerInsanity.
			return Insanity.InsanityDrain();
		end

		-- Insanity Drain Stacks
		self.DrainStacks = function()
			local Insanity = addonTable.Insanity;

			-- Stacks tracked by trackerInsanity.
			return Insanity.InsanityDrainStacks();
		end
	end

	-- power.max
	self.Max = function()
		local Tables = addon.Core.Tables;
		local apiUnit = addon.API.Unit;

		local powerInfo = apiUnit.GetPowerInfo(unitID, powerType);

		return Tables.GetValue(powerInfo, "Max", 1);
	end

	self.MaxCasts = function(Spell)
		local returnValue = 0;

		if Spell then
			local counter = 1;
			local powerAfterCast = self.AfterCast(Spell, counter);

			-- Stop checking after 10 casts is to allow for when a buff reduces the cost to zero as it would go into an infinite loop.
			while powerAfterCast > 0 and counter <= 10 do
				returnValue = counter;

				counter = counter + 1;
				powerAfterCast = self.AfterCast(Spell, counter);
			end
		end

		return returnValue;
	end

	-- power.pct
	self.Percent = function()
		return (self.Current() / self.Max()) * 100;
	end

	-- power.regen
	self.Regen = function()
		local Tables = addon.Core.Tables;
		local apiUnit = addon.API.Unit;

		local inCombat = apiUnit.InCombat();
		local powerInfo = apiUnit.GetPowerInfo(unitID, powerType);
		-- Truncate to 1 decimal place, truncate always rounds towards 0, should err on the side of caution and overestimate time to resources.
		local combatRegen = Tables.GetValue(powerInfo, "CombatRegen", 0);
		local noCombatRegen = Tables.GetValue(powerInfo, "NoCombatRegen", 0);

		-- These power types have natural regen without actions.
		if powerType == PowerType.Mana or powerType == PowerType.Focus or powerType == PowerType.Energy then
			return (inCombat and combatRegen)
				or noCombatRegen;
		-- These power types don't actually regen, though they can degenerate out of combat.
		else
			return (not inCombat and noCombatRegen)
				or 0;
		end

		return 0;
	end

	-- "power.regen.pct"
	self.RegenPercent = function()
		return (self.Regen() / self.Max()) * 100;
	end

	-- power.time_to_max
	self.TimeToMax = function()
		return self.TimeToX(self.Max());
	end

	self.TimeToNext = function()
		return self.TimeToX(min(self.Current() + 1, self.Max()));
	end

	-- "power.time_to_x"
	self.TimeToX = function(Amount)
		if powerType == PowerType.Runes then
			return ComputerTimeToXRunes(Amount);
		else
			if self.Regen() > 0 then
				return max((Amount - self.Current()) / self.Regen(), 0);
			else
				return huge;
			end
		end

		return 0;
	end

	-- "power.time_to_x.pct"
	self.TimeToXPercent = function(Percent)
		return self.TimeToX(self.Max() * (Percent / 100));
	end

	-- For keeping track of point type power types to track spending.
	if powerType == PowerType.ArcaneCharges or powerType == PowerType.Chi or powerType == PowerType.ComboPoints or powerType == PowerType.Runes or powerType == PowerType.SoulShards then
		local powerPoints = {
			["ARCANE_CHARGES"] = PowerType.ArcaneCharges,
			["Chi"] = PowerType.Chi,
			["COMBO_POINTS"] = PowerType.ComboPoints,
			["RUNES"] = PowerType.Runes,
			["SOUL_SHARDS"] = PowerType.SoulShards,
		};

		local prevPoints = 0;
		local currentPoints = 0;

		self.Previous = function()
			return prevPoints or 0;
		end

		local function powerUpdate(_, unitID, powerName)
			local coreTables = addon.Core.Tables;
			local power = coreTables.GetValue(powerPoints, powerName, -1);

			if unitID == "player" and power == powerType then
				local newPoints = self.Current();

				if newPoints ~= currentPoints then
					prevPoints = currentPoints;
					currentPoints = newPoints;
				end
			end
		end

		local tableEvents = {
			UNIT_POWER_UPDATE = powerUpdate,
		};

		addon.System.Events.RegisterEvents(tableEvents, true);
	end

	setmetatable(self, { __call = function(table, ...) return self.Current(...); end });
	Cache[key] = self;

	return self;
end