-- Pulls back the Addon-Local Variables and store them locally.
local addonName, addonTable = ...;
local addon = _G[addonName];
local L = LibStub("AceLocale-3.0"):GetLocale(addonName);

-- Store local copy of Blizzard API wrapper and other addon global functions and variables
local Blizzard, Callbacks, Events = addon.API.Blizzard, addon.System.Callbacks, addon.System.Events;
local error, format, setmetatable, sort, tinsert, tostring, type, wipe = error, format, setmetatable, sort, tinsert, tostring, type, wipe;

local rotationIndex = 0;
local actions = {};
local evaluateActions = {};

local function evalutateCheckBox(checkbox)
	local Settings = addon.Core.Settings;

	return (checkbox == nil and 2)
		or (type(checkbox) == "boolean" and checkbox == true and 2)
		or (type(checkbox) == "boolean" and checkbox == false and 1)
		or Settings.GetCharacterValue("ScriptOptions", checkbox)
		or Settings.GetRotationValue(checkbox);
end

local function sortActions(a, b)
	if a.WaitTime == b.WaitTime then
		return a.RotationIndex < b.RotationIndex;
	end

	return a.WaitTime < b.WaitTime;
end

local function validObjects(Object, useObject)
	local Objects = addon.Core.Objects;

	return (Objects.IsSpell(Object) and Object.IsKnown())
	   and (useObject == true or type(useObject) == "function");
end

local Prototype = {
	Actions = function()
		return actions;
	end,

	EvaluateDefensiveAction = function(Object, useObject, checkbox)
		rotationIndex = rotationIndex + 1;

		if validObjects(Object, useObject) then
			local Tables = addon.Core.Tables;
			local evaluateAction = Object.Evaluate();

			if Tables.GetValue(evaluateAction, "Use", false) and evalutateCheckBox(checkbox or Object.CheckBox()) >= 1
															 and Object.Charges() >= 1
									   						 and addon.currentRotation.ProcessCondition(useObject) then
				local spellID = Object.SpellID();
				local objectInfo = addon.API.Spell.GetSpellInfo(spellID);

				if not Tables.IsEmpty(objectInfo) then
					local iconInfo = {
						ID = spellID,
						Name = objectInfo.Name,
						Texture = objectInfo.IconID,
						Text = "",
						Type = "Spell",
						Keybind = addon.Modules.ActionBar.Keybind("Spell", spellID),
						Wait = false,
						WaitStart = 0,
						WaitEnd = 0,
						WaitTime = 0;
						RotationIndex = rotationIndex,
					};

					tinsert(evaluateActions, iconInfo);
				end
			end
		end
	end,
};

local ClassMT = {
	__index = Prototype,

	__newindex = function(table, key, value)
		error(format("Attempted to overwrite protected function %s in the Defensive Icon object!", key));
	end,
};

addonTable.iconDefensive = setmetatable({}, ClassMT);

local function timerEvent()
	local Debug = addonTable.Debug;
	local Tables = addon.Core.Tables;
	local Target = addon.Units.Target;

	local currentRotation = addon.currentRotation;
	local iconDefensive = addonTable.iconDefensive;

	if not Debug then
		if Target.IsValidTarget() and currentRotation then
			rotationIndex = 0;
			evaluateActions = {};

			local action = Tables.GetValue(actions, 1, {});
			local actionID = format("%s-%s-s", Tables.GetValue(action, "Type", ""), Tables.GetValue(action, "ID", ""), tostring(Tables.GetValue(action, "Wait", "")));

			wipe(actions);

			currentRotation.ProcessDefensive(iconDefensive);
			sort(evaluateActions, sortActions);

			actions = Tables.Duplicate(evaluateActions);
			action = Tables.GetValue(actions, 1, {});

			if format("%s-%s-%s", Tables.GetValue(action, "Type", ""), Tables.GetValue(action, "ID", ""), tostring(Tables.GetValue(action, "Wait", ""))) ~= actionID then
				Callbacks.Callback("DefensiveIcon", actions);
			end

			evaluateActions = nil;
		else
			wipe(actions);

			Callbacks.Callback("DefensiveIcon");
		end
	end
end

local function enterCombat()
	local systemTimers = addon.System.Timers;
	local currentRotation = addon.currentRotation;

	if currentRotation and not systemTimers.TimerExists(timerEvent) then
		timerEvent();
		systemTimers.TimerAdd(timerEvent, 1);
	end
end

local function leaveCombat()
	local systemTimers = addon.System.Timers;

	systemTimers.TimerRemove(timerEvent);
	wipe(actions);

	Callbacks.Callback("DefensiveIcon");
end

local function targetChanged()
	local apiUnit = addon.API.Unit;
	local Target = addon.Units.Target;

	if Target.IsValidTarget(true) then
		enterCombat();
	elseif not apiUnit.InCombat() then
		leaveCombat();
	end
end

local moduleEvents = {
	Disabled = function()
		leaveCombat();
	end,

	Enabled = function()
		local currentRotation = addon.currentRotation;
		local apiUnit = addon.API.Unit;

		if currentRotation and apiUnit.InCombat() then
			enterCombat();
		end
	end,

	unitDied = function(unitID)
		local Unit = addon.API.Unit;

		if unitID == "target" and not Unit.InCombat() then
			leaveCombat();
		end
	end,
};

local tableEvents = {
	INSTANCE_ENCOUNTER_ENGAGE_UNIT = enterCombat,
	PLAYER_REGEN_DISABLED = enterCombat,
	PLAYER_REGEN_ENABLED = leaveCombat,
	PLAYER_TARGET_CHANGED = targetChanged,
};

Callbacks.RegisterAddonCallbacks(moduleEvents);
Events.RegisterEvents(tableEvents);