-- Pulls back the Addon-Local Variables and store them locally.
local addonName, addonTable = ...;
local addon = _G[addonName];
local L = LibStub("AceLocale-3.0"):GetLocale(addonName);

-- Store local copy of Blizzard API wrapper and other addon global functions and variables
local Blizzard, systemCallbacks, systemEvents = addon.API.Blizzard, addon.System.Callbacks, addon.System.Events;
local debugprofilestop, format, max, pairs, setmetatable, sort, tinsert, tostring, wipe = debugprofilestop, format, max, pairs, setmetatable, sort, tinsert, tostring, wipe;

local eventFrame = addon.API.Blizzard.CreateFrame("Frame");

local enableProcessing = false;
local lastRefresh = 0;

local Evaluate, Events = {}, {};

Evaluate = {
	Cooldown = function(evaluateAction)
		local Player = addon.Units.Player;
		local Spell = addon.API.Spell;
		local Tables = addon.Core.Tables;

		if not Tables.IsEmpty(evaluateAction) then
			local castingEnd = Player.Casting.End();
			local currentTime = addon.API.General.GetTime();
			local endTime = Tables.GetValue(evaluateAction, "End", 0);
			local GCDInfo = Spell.GetSpellCooldown(addonTable.GLOBAL_COOLDOWN);
			local GCDEnd = Tables.GetValue(GCDInfo, "End", 0);

			if endTime > max(GCDEnd, castingEnd) then
				return Tables.GetValue(evaluateAction, "Start", currentTime), endTime;
			elseif castingEnd > max(endTime, GCDEnd) then
				return Player.Casting.Start("player") or currentTime, castingEnd;
			elseif GCDEnd > max(endTime, castingEnd) then
				return Tables.GetValue(GCDInfo, "Start", currentTime), GCDEnd;
			end
		end

		return 0, 0;
	end,

	OutOfRange = function(evaluateNoRange, evaluateRange)
		local Tables = addon.Core.Tables;

		if not Tables.IsEmpty(evaluateNoRange) then
			if not Tables.IsEmpty(evaluateRange) then
				local firstNoRange = Tables.GetValue(evaluateNoRange, 1);
				local firstRange = Tables.GetValue(evaluateRange, 1);

				return Tables.GetValue(firstNoRange, "WaitTime") < Tables.GetValue(firstRange, "WaitTime");
			else
				return true;
			end
		end

		return false;
	end,

	Rotation = function(forceRefresh)
		local currentTime = debugprofilestop();

		if (forceRefresh or currentTime - lastRefresh >= 100) and not addonTable.Debug then
			local Profiling = addon.Core.Profiling.New("iconRotation.timerEvent", 1);
			local Rotation = addon.Rotation;
			local Target = addon.Units.Target;

			lastRefresh = currentTime;
			addonTable.accountSettings.Profiling = addonTable.accountSettings.Profiling or {};

			local rotationIcons = { SingleTargetIcon = addon.Icons.ST, MultiTargetIcon = addon.Icons.AOE, OffGCDIcon = addon.Icons.OFFGCD, CooldownIcon = addon.Icons.CD };

			if addon.currentRotation and Target.IsValidTarget() then
				Profiling.Start();
				Rotation.StartRotation();

				addon.MinDistance = Target.InRange.Min();
				addon.MaxDistance = Target.InRange.Max();

				for _, rotationIcon in pairs(rotationIcons) do
					rotationIcon.StartRotation();
				end

				addon.currentRotation.ProcessRotation();

				for _, rotationIcon in pairs(rotationIcons) do
					rotationIcon.EndRotation();
				end

				Profiling.Stop();
			else
				for iconName, rotationIcon in pairs(rotationIcons) do
					wipe(rotationIcon.Actions());
					systemCallbacks.Callback(iconName);
				end
			end
		end
	end,
};

local function sortActions(a, b)
	local coreTables = addon.Core.Tables;

	if coreTables.GetValue(a, "WaitTime", 0) == coreTables.GetValue(b, "WaitTime", 0) then
		return coreTables.GetValue(a, "RotationIndex", 0) < coreTables.GetValue(b, "RotationIndex", 0);
	end

	return coreTables.GetValue(a, "WaitTime", 0) < coreTables.GetValue(b, "WaitTime", 0);
end

local Cache = setmetatable({}, { __mode = "kv", __metatable = "This table is protected!" });

addonTable.iconRotation = function(ID)
	local self = Cache[ID] or {};

	local actions = {};

	local evaluateNoRange = {};
	local evaluateRange = {};

	local iconName = ID;
	local rotationIndex = 0;
	local rotationPaused = false;
	local stopProcessing = false;

	local function evaluateIcon(rotationIndex, Object, iconText, evaluateAction)
		local ActionBar = addon.Modules.ActionBar;
		local Item = addon.API.Item;
		local Objects = addon.Core.Objects;
		local Settings = addon.Core.Settings;
		local Spell = addon.API.Spell;
		local Tables = addon.Core.Tables;

		local useFullProcessing = Settings.GetAccountValue("AddonSettings", "useFullProcessing") == 1;
		local ObjectID = Objects.IsSpell(Object) and Object.SpellID() or Object.ItemID();
		local objectInfo = Objects.IsSpell(Object) and Spell.GetSpellInfo(ObjectID) or Item.GetItemInfo(ObjectID);
		local ObjectType = Objects.GetType(Object);
		local startTime, endTime = Evaluate.Cooldown(evaluateAction);

		if not Tables.IsEmpty(objectInfo) then
			local iconInfo = {
				ID = ObjectID,
				Name = objectInfo.Name,
				Texture = objectInfo.IconID,
				Text = iconText or "",
				Type = ObjectType,
				Keybind = ActionBar.Keybind(ObjectType, ObjectID),
				Wait = endTime > Blizzard.GetTime(),
				WaitStart = startTime,
				WaitEnd = endTime,
				WaitTime = endTime - startTime;
				RotationIndex = rotationIndex,
			};

			if Object.InRange(addon.Units.Target) then
				tinsert(evaluateRange, iconInfo);
			end

			tinsert(evaluateNoRange, iconInfo);

			-- Action can be used now, no waiting for resources or cooldowns, set the rotation to stop processing.
			if not useFullProcessing and Tables.GetValue(evaluateAction, "End", -1) == 0 then
				stopProcessing = true;
			end
		end
	end

	local function outOfRange()
		if (iconName == "SingleTarget" or iconName == "MultiTarget") and Evaluate.OutOfRange(evaluateNoRange, evaluateRange) then
			rotationIndex = (rotationIndex or 0) + 1;

			local iconInfo = {
				ID = 0,
				Name = "Out of Range",
				Texture = 450907,
				Text = "move",
				Type = "Special",
				Keybind = "",
				Wait = false,
				WaitStart = 0,
				WaitEnd = 0,
				WaitTime = 0;
				RotationIndex = rotationIndex,
			};

			-- We are supposed to show out of range so put it first in the queue, this saves needing to resort again.
			tinsert(evaluateRange, 1, iconInfo);
		end
	end

	self.Actions = function()
		return actions;
	end

	self.EndRotation = function()
		local Settings = addon.Core.Settings;
		local Tables = addon.Core.Tables;

		local action = Tables.GetValue(actions, 1, {});
		local actionID = format("%s-%s-s", Tables.GetValue(action, "Type", ""), Tables.GetValue(action, "ID", ""), tostring(Tables.GetValue(action, "Wait", "")));

		wipe(actions);

		if Settings.GetCharacterValue("ScriptOptions", "OPT_RANGE_CHECK") == 1 then
			sort(evaluateNoRange, sortActions);
			sort(evaluateRange, sortActions);

			outOfRange();

			actions = Tables.Duplicate(evaluateRange);
		else
			sort(evaluateNoRange, sortActions);

			actions = Tables.Duplicate(evaluateNoRange);
		end

		action = Tables.GetValue(actions, 1, {});

		if format("%s-%s-%s", Tables.GetValue(action, "Type", ""), Tables.GetValue(action, "ID", ""), tostring(Tables.GetValue(action, "Wait", ""))) ~= actionID then
			systemCallbacks.Callback(format("%sIcon", iconName), actions);
		end

		wipe(evaluateNoRange);
		wipe(evaluateRange);
	end

	self.EvaluateAction = function(rotationIndex, Object, iconText, evaluateAction)
		evaluateIcon(rotationIndex, Object, iconText or addon.Core.Tables.GetValue(evaluateAction, "Text", ""), evaluateAction);
	end

	self.EvaluateWaitCondition = function(rotationIndex, Object)
		evaluateIcon(rotationIndex, Object, "wait");
	end

	self.Paused = function()
		return rotationPaused == true;
	end

	self.StartRotation = function()
		rotationIndex = 0;

		rotationPaused = false;
		stopProcessing = false;

		wipe(evaluateNoRange);
		wipe(evaluateRange);
	end

	self.Stopped = function(value)
		if value then
			stopProcessing = value;
		else
			return stopProcessing == true;
		end
	end

	Cache[ID] = self;

	return self;
end

Events = {
	Casting = function(...)
		if enableProcessing then
			local _, unitID = ...;

			if unitID == "player" or unitID == "pet" then
				Evaluate.Rotation();
			end
		end
	end,

	CastSuccess = function(...)
		if enableProcessing then
			local _, unitID = ...;

			if unitID == "player" or unitID == "pet" then
				Evaluate.Rotation(true);
			end
		end
	end,

	DisableProcessing = function()
		local rotationIcons = {
			SingleTargetIcon = addon.Icons.ST,
			MultiTargetIcon = addon.Icons.AOE,
			OffGCDIcon = addon.Icons.OFFGCD,
			CooldownIcon = addon.Icons.CD,
		};

		eventFrame:SetScript("OnUpdate", nil);
		enableProcessing = false;

		for iconName, rotationIcon in pairs(rotationIcons) do
			wipe(rotationIcon.Actions());
			systemCallbacks.Callback(iconName);
		end
	end,

	EnableProcessing = function()
		enableProcessing = true;
		eventFrame:SetScript("OnUpdate", Events.OnUpdate);

		if addon.currentRotation then
			-- Ensure Precombat Variables are run at least once to initialize.
			addon.currentRotation.PrecombatVariables(addonTable.RotationProcessing);
		end

		Evaluate.Rotation();
	end,

	OnUpdate = function()
		if enableProcessing then
			local currentTime = debugprofilestop();

			-- Run the loop only when there is something to update
			if currentTime - lastRefresh >= (addon.refreshInterval * 1000) and not addonTable.Debug then
				Evaluate.Rotation(true);
			end
		end
	end,

	Resources = function(...)
		if enableProcessing then
			local _, unitID = ...;

			if unitID == "player" or unitID == "pet" then
				Evaluate.Rotation();
			end
		end
	end,

	SpellUsable = function(...)
		if enableProcessing then
			Evaluate.Rotation();
		end
	end,

	TargetChange = function()
		if addon.Units.Target.IsValidTarget(true) then
			Events.EnableProcessing();
		elseif not addon.API.Unit.InCombat() then
			Events.DisableProcessing();
		end
	end,
};

local moduleEvents = {
	aurasChanged = function()
		if enableProcessing then
			Evaluate.Rotation();
		end
	end,

	Disabled = function()
		Events.DisableProcessing();
	end,

	Enabled = function()
		local Unit = addon.API.Unit;

		if Unit.InCombat() then
			Events.EnableProcessing();
		else
			Events.DisableProcessing();
		end
	end,

	enemiesChanged = function()
		Evaluate.Rotation();
	end,

	spellsChanged = function()
		Evaluate.Rotation();
	end,

	unitDied = function(unitID)
		local Unit = addon.API.Unit;

		if unitID == "target" and not Unit.InCombat() then
			Events.DisableProcessing();
		end
	end,
};

local tableEvents = {
	-- Instance Boss Fight Start
	INSTANCE_ENCOUNTER_ENGAGE_UNIT = Events.EnableProcessing,

	-- Combat Changed
	PLAYER_REGEN_DISABLED = Events.EnableProcessing,
	PLAYER_REGEN_ENABLED = Events.DisableProcessing,

	-- Target Changed
	PLAYER_TARGET_CHANGED = Events.TargetChange,

	-- Spell Usable/Unsuable Changed
	SPELL_UPDATE_USABLE = Events.SpellUsable,

	-- Resources Changed
	UNIT_POWER_UPDATE = Events.Resources,

	-- Cast Started
	UNIT_SPELLCAST_CHANNEL_START = Events.Casting,
	UNIT_SPELLCAST_START = Events.Casting,

	-- Cast Stopped
	UNIT_SPELLCAST_CHANNEL_STOP = Events.Casting,
	UNIT_SPELLCAST_FAILED = Events.Casting,
	UNIT_SPELLCAST_INTERRUPTED = Events.Casting,
	UNIT_SPELLCAST_STOP = Events.Casting,

	-- Cast Succeeded
	UNIT_SPELLCAST_SUCCEEDED = Events.CastSuccess,
};

systemCallbacks.RegisterAddonCallbacks(moduleEvents);
systemEvents.RegisterEvents(tableEvents);