-- Pulls back the Addon-Local Variables and store them locally.
local addonName, addonTable = ...;
local addon = _G[addonName];
local L = LibStub("AceLocale-3.0"):GetLocale(addonName);

local StaticPopupDialogs = StaticPopupDialogs;

StaticPopupDialogs["LERH_FULL_PROCESSING_DISABLE"] = {
	text = "Disabling full rotation processing will increase your FPS, however you will no longer be able to use the Hekili Weak Aura Style.\n\nIf you click 'Yes' the Hekili Weak Aura Style will be removed, and your Weak Aura Style will be reset to Ovale if you currently have Hekili Style chosen!\n\nThe UI must be reloaded if you click 'Yes'!\n\n",
	button1	= "Yes",
	button2	= "No",
	OnAccept = function()
		addon.Core.Settings.SetAccountValue("AddonSettings", "useFullProcessing", 0);
		addon.Features.WeakAuras.VerifyStyle();
		addon.API.Blizzard.ReloadUI();
	end,
	whileDead = true,
	hideOnEscape = true,
};
StaticPopupDialogs["LERH_FULL_PROCESSING_ENABLE"] = {
	text = "WARNING: Enabling full rotation processing provides no benefit besides making the Hekili Weak Aura Style usable and WILL DROP YOUR FPS because of the extra processing of actions.\n\nIf you don't plan on using the Hekili Weak Aura Style, then click the 'No' button!\n\nThe UI must be reloaded if you click 'Yes'!\n\n",
	button1	= "Yes",
	button2	= "No",
	OnAccept = function()
		addon.Core.Settings.SetAccountValue("AddonSettings", "useFullProcessing", 1);
		addon.API.Blizzard.ReloadUI();
	end,
	whileDead = true,
	hideOnEscape = true,
};

local function getRotationName()
	local pairs = pairs;
	local Tables = addon.Core.Tables;
	local apiGeneral = addon.API.General;

	local specInfo = apiGeneral.GetSpecializationInfo();

	if not Tables.IsEmpty(specInfo) then
		local numRotations, rotationList = addon.Rotation.GetRegisteredRotations(specInfo.ID);

		if numRotations == 1 then
			for key in pairs(rotationList) do
				return key;
			end
		else
			return "__none";
		end
	end
end

addon.Options.Core = {
	CleanupOptions = function(optionsTable)
		local pairs = pairs;
		local Tables = addon.Core.Tables;

		for key, value in pairs(optionsTable.args) do
			if Tables.IsEmpty(value.args) then
				optionsTable.args[key] = nil;
			end
		end

		return optionsTable;
	end,

	ConvertToValue = function(value, tristate)
		if value == nil then
			return 1;
		elseif value then
			return (tristate and 2) or 1;
		else
			return 0;
		end
	end,

	ConvertToUI = function(value, tristate)
		if value == 2 or (not tristate and value == 1) then
			return true;
		elseif value == 0 then
			return false;
		else
			return nil;
		end
	end,

	CreaterHeader = function(headerText)
		local format = format;

		return format("%s%s%s", addonTable.COLORS.TAGS.HIGHLIGHT, headerText, addonTable.TAGS.END);
	end,
};

addon.Options.Settings = {
	GetAccountValue = function(info)
		local Tables = addon.Core.Tables;
		local returnValue = true;

		if info then
			local sectionName = info[#info - 1];
			local keyName = info[#info];
			local profile = Tables.GetValue(addonTable.accountSettings, sectionName, {});

			returnValue = Tables.GetValue(profile, keyName);

			if info.option.type == "toggle" then
				returnValue = addon.Options.Core.ConvertToUI(returnValue, info.option.tristate);
			end
		end

		return returnValue;
	end,

	SetAccountValue = function(info, value)
		if info then
			local Tables = addon.Core.Tables;

			if info.option.type == "toggle" then
				value = addon.Options.Core.ConvertToValue(value, info.option.tristate);
			end

			local sectionName = info[#info - 1];
			local keyName = info[#info];
			local profile = Tables.GetValue(addonTable.accountSettings, sectionName, {});

			if keyName == "refreshInterval" then
				addon.refreshInterval = value / 1000;
			elseif keyName == "reactTime" then
				addon.reactTime = value / 1000;
			elseif keyName == "useFullProcessing" then
				-- Changed the rotation processing mode, show the popup dialog box.
				if value == 1 then
					addon.API.Blizzard.StaticPopup_Show("LERH_FULL_PROCESSING_ENABLE");
				else
					addon.API.Blizzard.StaticPopup_Show("LERH_FULL_PROCESSING_DISABLE");
				end
			end

			profile[keyName] = value;

			-- Put the settings into the database
			addonTable.accountSettings[sectionName] = profile;
		end
	end,

	GetActiveRotation = function(info)
		local Tables = addon.Core.Tables;
		local apiGeneral = addon.API.General;

		local returnValue;

		if info then
			local specInfo = apiGeneral.GetSpecializationInfo();

			if not Tables.IsEmpty(specInfo) then
				local profile = Tables.GetValue(addonTable.characterSettings, "ActiveRotations", {});

				returnValue = Tables.GetValue(profile, specInfo.Name);

				if not returnValue then
					returnValue = getRotationName();
					profile[specInfo.Name] = returnValue;

					-- Put the settings into the database
					addonTable.characterSettings.ActiveRotations = profile;
				end
			end
		end

		return returnValue;
	end,

	SetActiveRotation = function(info, value)
		if info then
			local Rotation = addon.Rotation;

			Rotation.SetActiveRotation(value);
		end
	end,

	GetCharacterValue = function(info)
		local Tables = addon.Core.Tables;
		local returnValue = true;

		if info then
			local sectionName = info[#info - 1];
			local keyName = info[#info];
			local profile = Tables.GetValue(addonTable.characterSettings, sectionName, {});

			returnValue = Tables.GetValue(profile, keyName);

			if info.option.type == "toggle" then
				returnValue = addon.Options.Core.ConvertToUI(returnValue, info.option.tristate);
			end
		end

		return returnValue;
	end,

	SetCharacterValue = function(info, value)
		if info then
			local Tables = addon.Core.Tables;

			if info.option.type == "toggle" then
				value = addon.Options.Core.ConvertToValue(value, info.option.tristate);
			end

			local sectionName = info[#info - 1];
			local keyName = info[#info];
			local profile = Tables.GetValue(addonTable.characterSettings, sectionName, {});

			profile[keyName] = value;

			-- Put the settings into the database
			addonTable.characterSettings[sectionName] = profile;
		end
	end,

	GetNameplatesValue = function(info)
		local Tables = addon.Core.Tables;
		local returnValue = true;

		if info then
			local keyName = info[#info];
			local profile = Tables.GetValue(addonTable.accountSettings, "NameplateSettings", {});

			returnValue = Tables.GetValue(profile, keyName);

			if info.option.type == "toggle" then
				returnValue = addon.Options.Core.ConvertToUI(returnValue, info.option.tristate);
			end
		end

		return returnValue;
	end,

	SetNameplatesValue = function(info, value)
		if info then
			local Tables = addon.Core.Tables;

			if info.option.type == "toggle" then
				value = addon.Options.Core.ConvertToValue(value, info.option.tristate);
			end

			local keyName = info[#info];
			local profile = Tables.GetValue(addonTable.accountSettings, "NameplateSettings", {});

			profile[keyName] = value;

			if keyName ~= "nameplateShowAlways" then
				addon.API.Blizzard.SetCVar(keyName, profile[keyName]);
			end

			-- Put the settings into the database
			addonTable.accountSettings.NameplateSettings = profile;
		end
	end,

	GetPluginValue = function(info)
		local Tables = addon.Core.Tables;
		local returnValue = false;

		if info then
			local keyName = info[#info];
			local profile = Tables.GetValue(addonTable.characterSettings, "Plugins", {});

			returnValue = Tables.GetValue(profile, keyName);

			if info.option.type == "toggle" then
				returnValue = addon.Options.Core.ConvertToUI(returnValue, info.option.tristate);
			end
		end

		return returnValue;
	end,

	SetPluginValue = function(info, value)
		if info then
			local Tables = addon.Core.Tables;

			if info.option.type == "toggle" then
				value = addon.Options.Core.ConvertToValue(value, info.option.tristate);
			end

			local keyName = info[#info];
			local profile = Tables.GetValue(addonTable.characterSettings, "Plugins", {});

			profile[keyName] = value;

			-- Put the settings into the database
			addonTable.characterSettings.Plugins = profile;
		end
	end,

	GetRotationValue = function(info)
		local Tables = addon.Core.Tables;
		local apiGeneral = addon.API.General;

		local returnValue = true;

		if info then
			local specInfo = apiGeneral.GetSpecializationInfo();

			if not Tables.IsEmpty(specInfo) then
				local keyName = info[#info];
				local profile = Tables.GetValue(addonTable.characterSettings.RotationSettings, specInfo.Name, {});

				returnValue = Tables.GetValue(profile, keyName);

				if info.option.type == "toggle" then
					if returnValue == nil then
						returnValue = (info.option.tristate and 2) or 1;

						profile[keyName] = returnValue;

						-- Put the settings into the database
						addonTable.characterSettings.RotationSettings[specInfo.Name] = profile;
					else
						returnValue = addon.Options.Core.ConvertToUI(returnValue, info.option.tristate);
					end
				else
					returnValue = Tables.GetValue(profile, keyName);
				end
			end
		end

		return returnValue;
	end,

	SetRotationValue = function(info, value)
		if info then
			local Tables = addon.Core.Tables;
			local apiGeneral = addon.API.General;

			if info.option.type == "toggle" then
				value = addon.Options.Core.ConvertToValue(value, info.option.tristate);
			end

			local specInfo = apiGeneral.GetSpecializationInfo();

			if not Tables.IsEmpty(specInfo) then
				local keyName = info[#info];
				local profile = Tables.GetValue(addonTable.characterSettings.RotationSettings, specInfo.Name, {});

				profile[keyName] = value;

				-- Put the settings into the database
				addonTable.characterSettings.RotationSettings[specInfo.Name] = profile;
			end
		end
	end,
};
