-- Pulls back the Addon-Local Variables and store them locally.
local addonName, addonTable = ...;
local addon = _G[addonName];
local L = LibStub("AceLocale-3.0"):GetLocale(addonName);

-- Store local copy of Blizzard API wrapper and other addon global functions and variables
local Blizzard, systemEvents, systemTimers = addon.API.Blizzard, addon.System.Events, addon.System.Timers;
local ipairs, type = ipairs, type;

-- List of CLEU events for applying auras
local CLEU_SPELL_CAST_START = {
	SPELL_CAST_START = true,
	SPELL_CAST_SUCCESS = true,
};

local CLEU_SPELL_CAST_FAIL = {
	SPELL_CAST_FAILED = true,
};

local CLEU_SPELL_HIT = {
	SPELL_DAMAGE = true,
	SPELL_HEAL = true,
	SPELL_MISSED = true,
};

local PredictionCache = {};
local Events, Prediction = {}, {};

local function getInfo(auraData)
	if type(auraData) == "table" then
		return auraData[1], auraData[2];
	elseif type(auraData) == "number" then
		return auraData, nil;
	end
end

Prediction = {
	ConsumeAura = function(GUID, spellID, value)
		local Tables = addon.Core.Tables;

		if GUID and spellID then
			local consumeAuraData = Tables.GetValue(addonTable.ConsumeAuraData, spellID);

			if consumeAuraData then
				local spellIDs, requiredAura = getInfo(consumeAuraData);

				if spellIDs and (requiredAura == nil or (requiredAura and Prediction.GetAura(GUID, requiredAura))) then
					local aura, spellName;

					if not Tables.IsEmpty(spellIDs) then
						for _, ID in ipairs(spellIDs) do
							spellName = Blizzard.GetSpellInfo(ID);
							aura = Prediction.GetAura(GUID, ID);

							if aura then
								aura.Consumed = value;
								break;
							end
						end
					else
						spellName = Blizzard.GetSpellInfo(spellIDs);
						aura = Prediction.GetAura(GUID, spellIDs);

						if aura then
							aura.Consumed = value;
						end
					end
				end
			end
		end
	end,

	CreateAura = function(GUID, spellID)
		local Tables = addon.Core.Tables;

		if GUID and spellID then
			local unitAuras = PredictionCache[GUID] or {};
			local createAuraData = Tables.GetValue(addonTable.CreateAuraData, spellID);

			if createAuraData then
				local spellIDs, requiredAura = getInfo(createAuraData);

				if spellIDs and (requiredAura == nil or (requiredAura and Prediction.GetAura(GUID, requiredAura))) then
					local aura, spellName;

					if not Tables.IsEmpty(spellIDs) then
						for _, ID in ipairs(spellIDs) do
							spellName = Blizzard.GetSpellInfo(ID);
							aura = Prediction.GetAura(GUID, ID);

							if not aura then
								unitAuras[ID] = spellID;
								break;
							end
						end
					else
						spellName = Blizzard.GetSpellInfo(spellIDs);
						aura = Prediction.GetAura(GUID, spellIDs);

						if not aura then
							unitAuras[spellIDs] = spellID;
						end
					end

					PredictionCache[GUID] = unitAuras;
				end
			end
		end
	end,

	GetAura = function(GUID, spellID)
		if GUID and spellID then
			return Prediction.GetBuff(GUID, spellID) or Prediction.GetDebuff(GUID, spellID);
		end
	end,

	GetBuff = function(GUID, spellID)
		local Auras = addonTable.Auras;
		local returnValue;

		if GUID and spellID then
			returnValue = Auras.UnitBuff(nil, spellID, nil, nil, GUID);
		end

		return returnValue;
	end,

	GetDebuff = function(GUID, spellID)
		local Auras = addonTable.Auras;
		local returnValue;

		if GUID and spellID then
			returnValue = Auras.UnitDebuff(nil, spellID, nil, nil, GUID);
		end

		return returnValue;
	end,

	RemoveAura = function(GUID, spellID)
		local Tables = addon.Core.Tables;

		if GUID and spellID then
			local unitAuras = PredictionCache[GUID] or {};
			local createAuraData = Tables.GetValue(addonTable.CreateAuraData, spellID);

			if createAuraData then
				local spellIDs, requiredAura = getInfo(createAuraData);

				if spellIDs and (requiredAura == nil or (requiredAura and Prediction.GetAura(GUID, requiredAura))) then
					local aura, spellName;

					if not Tables.IsEmpty(spellIDs) then
						for _, ID in ipairs(spellIDs) do
							spellName = Blizzard.GetSpellInfo(ID);
							aura = unitAuras[ID];

							if aura and aura == spellID then
								unitAuras[ID] = nil;
								break;
							end
						end
					else
						spellName = Blizzard.GetSpellInfo(spellIDs);
						aura = unitAuras[spellIDs];

						if aura and aura == spellID then
							unitAuras[spellIDs] = nil;
						end
					end

					PredictionCache[GUID] = unitAuras;
				end
			end
		end
	end,
};

Events = {
	combatLog = function(...)
		local _, _, combatEvent, _, sourceGUID, _, _, _, destGUID, _, _, _, spellID = ...;

		if sourceGUID == Blizzard.UnitGUID("player") then
			if CLEU_SPELL_CAST_START[combatEvent] then
				-- Spell cast started, mark any spells consumed, and add any fake temp auras.
				Prediction.ConsumeAura(sourceGUID, spellID, true);
				Prediction.CreateAura(sourceGUID, spellID);
			elseif CLEU_SPELL_CAST_FAIL[combatEvent] or CLEU_SPELL_HIT[combatEvent] then
				-- Spell cast either failed or hit meaning the auras should be up to date so remove temp changes.
				Prediction.ConsumeAura(sourceGUID, spellID);
				Prediction.RemoveAura(sourceGUID, spellID);
			end
		end
	end,
};

addonTable.Prediction.UnitAura = function(unitID, spellID)
	local GUID = Blizzard.UnitGUID(unitID);

	if GUID and spellID then
		return addon.Core.Tables.GetValue(PredictionCache[GUID], spellID);
	end
end

local tableEvents = {
	COMBAT_LOG_EVENT_UNFILTERED = Events.combatLog,
};

systemEvents.RegisterEvents(tableEvents);