-- Pulls back the Addon-Local Variables and store them locally.
local addonName, addonTable = ...;
local addon = _G[addonName];
local L = LibStub("AceLocale-3.0"):GetLocale(addonName);

-- Store local copy of Blizzard API wrapper and other addon global functions and variables
local ipairs, max, unpack = ipairs, max, unpack;

-- CooldownTracking Variables
local cooldowns, recentCasts = {}, {};

-- Prediction Variables
local PredictionCharges, PredictionCooldown, PredictionKeybind, PredictionReplacements = {}, {}, {}, {};
local isChanneling = false;

-- Variables for the sub-sections functions
local Events, Prediction, ReplacementSpell = {}, {}, {};

Prediction = {
	CastEnd = function(spellID)
		PredictionCharges[spellID] = nil;
		PredictionCooldown[spellID] = nil;
	end,

	CastStart = function(currentSpell, spellID)
		PredictionCharges[spellID] = max(currentSpell.Charges() - 1, 0);
		PredictionCooldown[spellID] = currentSpell.Cooldown.Base() > 0;
	end,
};

ReplacementSpell = {
	CastEnd = function(classID, spellID)
		local Tables = addon.Core.Tables;
		local replacementSpellData = Tables.GetValue(addonTable.ReplacementSpellsData, classID);

		if replacementSpellData then
			if not Tables.IsEmpty(replacementSpellData) then
				local originalID, replacementID;

				for _, spellData in ipairs(replacementSpellData) do
					originalID, replacementID = unpack(spellData);

					PredictionReplacements[originalID] = nil;
					PredictionReplacements[replacementID] = nil;
					PredictionKeybind[replacementID] = nil;
				end
			else
				PredictionReplacements[classID] = nil;
				PredictionReplacements[replacementSpellData] = nil;
				PredictionKeybind[replacementSpellData] = nil;
			end
		end
	end,

	CastStart = function(classID, spellID)
		local moduleActionBar = addon.Modules.ActionBar;
		local Tables = addon.Core.Tables;

		local replacementSpellData = Tables.GetValue(addonTable.ReplacementSpellsData, classID);

		if replacementSpellData then
			if not Tables.IsEmpty(replacementSpellData) then
				local originalID, replacementID;

				for _, spellData in ipairs(replacementSpellData) do
					originalID, replacementID = unpack(spellData);

					PredictionReplacements[originalID] = false;
					PredictionReplacements[replacementID] = true;
					PredictionKeybind[replacementID] = moduleActionBar.Keybind("Spell", spellID);
				end
			else
				PredictionReplacements[classID] = false;
				PredictionReplacements[replacementSpellData] = true;
				PredictionKeybind[replacementSpellData] = moduleActionBar.Keybind("Spell", spellID);
			end
		end
	end,
};

Events = {
	CastEnd = function(...)
		local coreObjects = addon.Core.Objects;
		local _, unitID, _, spellID = ...;

		if unitID == "player" then
			local currentSpell = addonTable.getSpell(spellID);

			if coreObjects.IsSpell(currentSpell) then
				local classID = currentSpell.ClassID();

				isChanneling = false;

				Prediction.CastEnd(classID);
				ReplacementSpell.CastEnd(classID, currentSpell.SpellID());
			end
		end
	end,

	CastStart = function(...)
		local coreObjects = addon.Core.Objects;
		local _, unitID, _, spellID = ...;

		if unitID == "player" then
			local currentSpell = addonTable.getSpell(spellID);

			if coreObjects.IsSpell(currentSpell) then
				local classID = currentSpell.ClassID();

				Prediction.CastStart(currentSpell, classID);
				ReplacementSpell.CastStart(classID, currentSpell.SpellID());
			end
		end
	end,

	CastSuccess = function(...)
		local coreObjects = addon.Core.Objects;
		local _, unitID, _, spellID = ...;

		if unitID == "player" and not isChanneling then
			local currentSpell = addonTable.getSpell(spellID);

			if coreObjects.IsSpell(currentSpell) then
				local classID = currentSpell.ClassID();

				Prediction.CastEnd(classID);
				ReplacementSpell.CastEnd(classID, currentSpell.SpellID());
			end
		end
	end,

	ChannelStart = function(...)
		local coreObjects = addon.Core.Objects;
		local _, unitID, _, spellID = ...;

		if unitID == "player" then
			local currentSpell = addonTable.getSpell(spellID);

			if coreObjects.IsSpell(currentSpell) then
				local classID = currentSpell.ClassID();

				isChanneling = true;

				Prediction.CastStart(currentSpell, classID);
				ReplacementSpell.CastStart(classID, currentSpell.SpellID());
			end
		end
	end,
};

addonTable.Prediction.Charges = function(spellID)
	return PredictionCharges[spellID];
end

addonTable.Prediction.Cooldown = function(spellID)
	return PredictionCooldown[spellID];
end

addonTable.Prediction.SpellKeybind = function(spellID)
	return PredictionKeybind[spellID];
end

addonTable.Prediction.SpellKnown = function(spellID)
	return PredictionReplacements[spellID];
end

local tableEvents = {
	UNIT_SPELLCAST_STOP = Events.CastEnd, 						-- Fires when a unit stops or cancels casting a spell
	UNIT_SPELLCAST_CHANNEL_START = Events.ChannelStart, 		-- Fires when a unit starts channeling a spell
	UNIT_SPELLCAST_CHANNEL_STOP = Events.CastEnd, 				-- Fires when a unit stops or cancels a channeled spell
	UNIT_SPELLCAST_INTERRUPTED = Events.CastEnd, 				-- Fires when a unit's spell cast is interrupted
	UNIT_SPELLCAST_START = Events.CastStart, 					-- Fires when a unit begins casting a spell
	UNIT_SPELLCAST_SUCCEEDED = Events.CastSuccess, 				-- Fires when a unit's spell cast succeeds
};

addon.System.Events.RegisterEvents(tableEvents, true);