-- Pulls back the Addon-Local Variables and store them locally.
local _, addonTable = ...;
local addon = _G["LunaEclipse_Rotation"];

-- Plugin information variables.
local pluginName = "Equipment Manager";

-- Check whether the plugin is supposed to be active
if addon and addon.System.Plugins.IsActive(pluginName) then
    -- Store local copies of Blizzard API and other addon global functions and variables.
    local C_EquipmentSet = C_EquipmentSet;
    local ERR_LEARN_ABILITY_S, ERR_LEARN_PASSIVE_S, ERR_LEARN_SPELL_S, ERR_SPELL_UNLEARNED_S = ERR_LEARN_ABILITY_S, ERR_LEARN_PASSIVE_S, ERR_LEARN_SPELL_S, ERR_SPELL_UNLEARNED_S;
    local ipairs, pcall, string, tonumber = ipairs, pcall, string, tonumber;

    -- Variables for plugin.
    local equipSuccess = true;
    local prevItems;

    -- Tables to hold private functions.
    local ChatFilter, Events, Items, Set = {}, {}, {}, {};

    --------------------- Private Module Functions ---------------
    ChatFilter = {
        FilterMessage = function(_, _, msg, ...)
            if ChatFilter.Use(msg) then
                return true;
            end

            return false, msg, ...;
        end,

        Use = function(msg)
            return string.find(msg, string.gsub(ERR_LEARN_ABILITY_S:gsub("%.", "%."), "%%s", "(.*)"))
                or string.find(msg, string.gsub(ERR_LEARN_SPELL_S:gsub("%.", "%."), "%%s", "(.*)"))
                or string.find(msg, string.gsub(ERR_SPELL_UNLEARNED_S:gsub("%.", "%."), "%%s", "(.*)"))
                or string.find(msg, string.gsub(ERR_LEARN_PASSIVE_S:gsub("%.", "%."), "%%s", "(.*)"));
        end,
    };

    Items = {
        Current = function()
            local returnValue = {};

            -- No saved equipment set info so just get the current info.
            for counter = 1, 19 do
                local itemID = addon.API.Blizzard.GetInventoryItemID("player", counter);

                returnValue[counter] = itemID;
            end

            return returnValue;
        end,

        ItemIDs = function(setID)
            local success, returnValue = pcall(C_EquipmentSet.GetItemIDs, setID);

            if not success then
                local Tables = addon.Core.Tables;
                local setName = C_EquipmentSet.GetEquipmentSetInfo(setID);

                returnValue = addonTable.equipManager.Sets[setName];

                if Tables.IsEmpty(returnValue) then
                    returnValue = prevItems;
                end
            end

            return returnValue or {};
        end,
    };

    Set = {
        Equipped = function(ID)
            local Tables = addon.Core.Tables;
            local setInfo = Set.Info(ID);

            return Tables.GetValue(setInfo, "Equipped", false);
        end,

        Exists = function(ID)
            local setInfo = Set.Info(ID);

            return setInfo ~= nil;
        end,

        GetID = function(ID)
            local Tables = addon.Core.Tables;

            if Tables.IsTable(ID) then
                local equipmentSets = C_EquipmentSet.GetEquipmentSetIDs();

                if not Tables.IsEmpty(equipmentSets) then
                    for _, setID in ipairs(equipmentSets) do
                        local setName = C_EquipmentSet.GetEquipmentSetInfo(setID);

                        if setName == Tables.GetValue(ID, "Name", "") then
                            return setID;
                        end
                    end
                end
            else
                return tonumber(ID)
            end
        end,

        Icon = function(ID)
            local Tables = addon.Core.Tables;
            local setInfo = Set.Info(ID);

            return Tables.GetValue(setInfo, "Icon");
        end,

        ID = function(ID)
            local Tables = addon.Core.Tables;
            local setInfo = Set.Info(ID);

            return Tables.GetValue(setInfo, "ID");
        end,

        Info = function(ID)
            local equipmentSetID = Set.GetID(ID);
            local returnValue;

            if equipmentSetID then
                local setName, iconFileID, setID, isEquipped, numItems, numEquipped, numInventory, numLost, numIgnored = C_EquipmentSet.GetEquipmentSetInfo(equipmentSetID);
                local setLocations = C_EquipmentSet.GetItemLocations(setID);

                returnValue = {
                    Equipped = isEquipped,
                    Icon = iconFileID,
                    ID = setID,
                    ItemIDs = Items.ItemIDs(setID),
                    Locations = setLocations,
                    Name = setName,
                    numEquipped = numEquipped,
                    numIgnored = numIgnored,
                    numInventory = numInventory,
                    numItems = numItems,
                    numLost = numLost,
                };
            end

            return returnValue;
        end,

        ItemIDs = function(ID)
            local Tables = addon.Core.Tables;
            local setInfo = Set.Info(ID);

            return Tables.GetValue(setInfo, "ItemIDs", {});
        end,

        Load = function(specInfo)
            local Tables = addon.Core.Tables;

            if not Tables.IsEmpty(specInfo) and Set.Exists(specInfo) and not Set.Equipped(specInfo) then
                local setID = Set.ID(specInfo);

                addon.System.Events.EventRemove("PLAYER_EQUIPMENT_CHANGED", Events.EquipmentChanged, true);
                equipSuccess = C_EquipmentSet.UseEquipmentSet(setID);
            end
        end,

        Locations = function(ID)
            local Tables = addon.Core.Tables;
            local setInfo = Set.Info(ID);

            return Tables.GetValue(setInfo, "Locations", {});
        end,

        Name = function(ID)
            local Tables = addon.Core.Tables;
            local setInfo = Set.Info(ID);

            return Tables.GetValue(setInfo, "Name", "");
        end,

        NumberEquipped = function(ID)
            local Tables = addon.Core.Tables;
            local setInfo = Set.Info(ID);

            return Tables.GetValue(setInfo, "numEquipped", 0);
        end,

        NumberIgnored = function(ID)
            local Tables = addon.Core.Tables;
            local setInfo = Set.Info(ID);

            return Tables.GetValue(setInfo, "numIgnored", 0);
        end,

        NumberInventory = function(ID)
            local Tables = addon.Core.Tables;
            local setInfo = Set.Info(ID);

            return Tables.GetValue(setInfo, "numInventory", 0);
        end,

        NumberItems = function(ID)
            local Tables = addon.Core.Tables;
            local setInfo = Set.Info(ID);

            return Tables.GetValue(setInfo, "numItems", 0);
        end,

        NumberLost = function(ID)
            local Tables = addon.Core.Tables;
            local setInfo = Set.Info(ID);

            -- Default to 100 if there is no data so addon thinks all items are lost.
            return Tables.GetValue(setInfo, "numLost", 100);
        end,

        Save = function(specInfo)
            local Tables = addon.Core.Tables;

            if not Tables.IsEmpty(specInfo) and equipSuccess then
                local setID = Set.ID(specInfo);

                if Set.Exists(specInfo) then
                    C_EquipmentSet.SaveEquipmentSet(setID, specInfo.IconID);
                    C_EquipmentSet.AssignSpecToEquipmentSet(setID, specInfo.Index);
                else
                    C_EquipmentSet.CreateEquipmentSet(specInfo.Name, specInfo.IconID);

                    setID = Set.ID(specInfo);

                    C_EquipmentSet.AssignSpecToEquipmentSet(setID, specInfo.Index);
                end

                local setName = Set.Name(specInfo);

                addonTable.equipManager.Sets[setName] = Items.Current();
            end
        end,
    };

    ------------------------ WOW API Events ----------------------
    Events = {
        EquipmentChanged = function(...)
            local Tables = addon.Core.Tables;
            local apiGeneral = addon.API.General;
            local apiItem = addon.API.Item;
            local equipManager = addonTable.equipManager;

            local _, slotID = ...;
            local specInfo = apiGeneral.GetSpecializationInfo();
            local setItems = Set.ItemIDs(specInfo);
            local setItemID = Tables.GetValue(setItems, slotID, 0);

            if not Set.Exists(specInfo) or setItemID == 0 or Tables.GetValue(equipManager, "disableDowngrade", 1) == 0 then
                -- Either equipment set doesn't exist, or there was no item in the slot previously, or we are to save the set even if its a downgrade.
                addon.Features.EquipmentManager.Save(specInfo);
            else
                -- Don't save the equipment set if the item is a downgrade
                local itemID = addon.API.Blizzard.GetInventoryItemID("player", slotID);
                local currentItem = apiItem.GetItemInfo(itemID);
                local previousItem = apiItem.GetItemInfo(setItemID);
                local downgradeLevel = Tables.GetValue(equipManager, "downgradeLevel", 25);

                if Tables.GetValue(currentItem, "Level", 0) >= Tables.GetValue(previousItem, "Level", 0) - downgradeLevel then
                    -- Item is not considered a downgrade so save the equipment set.
                    addon.Features.EquipmentManager.Save(specInfo);
                end
            end
        end,

        SetChanged = function(...)
            local _, success = ...;
            local specInfo = addon.API.General.GetSpecializationInfo();

            if success and Set.Equipped(specInfo) then
                addon.System.Events.EventAdd("PLAYER_EQUIPMENT_CHANGED", Events.EquipmentChanged, true);
                prevItems = Items.Current();
            end
        end,
    };

    ---------------------- Public Functions ----------------------
    addon.Features.EquipmentManager = {
        Equipped = function(ID)
            return Set.Equipped(ID);
        end,

        Exists = function(ID)
            return Set.Exists(ID);
        end,

        Info = function(ID)
            return Set.Info(ID);
        end,

        ID = function(ID)
            return Set.ID(ID);
        end,

        Load = function(specInfo)
            addon.System.Events.CombatEventDelay(Set.Load, specInfo);
        end,

        Name = function(ID)
            return Set.Name(ID);
        end,

        Save = function(specInfo)
            Set.Save(specInfo);
        end,
    };

    ----------------------- Plugin Code Operation ----------------
    local moduleEvents = {
        playerLogin = function()
            addon.API.Blizzard.ChatFrame_AddMessageEventFilter("CHAT_MSG_SYSTEM", ChatFilter.FilterMessage);
            prevItems = Items.Current();
        end,

        playerLogout = function()
            addon.API.Blizzard.ChatFrame_RemoveMessageEventFilter("CHAT_MSG_SYSTEM", ChatFilter.FilterMessage);
        end,

        leaveCombat = function()
            local specInfo = addon.API.General.GetSpecializationInfo();

            if not equipSuccess and Set.NumberLost(specInfo) == 0 then
                addon.Features.EquipmentManager.Load(specInfo);
            end
        end,

        changeSpecStart = function(specInfo)
            addon.Features.EquipmentManager.Save(specInfo);
        end,

        changeSpecFinish = function(specInfo)
            if not Set.Exists(specInfo) then
                addon.Features.EquipmentManager.Save(specInfo);
            else
                addon.Features.EquipmentManager.Load(specInfo);
            end
        end,
    };

    local tableEvents = {
        EQUIPMENT_SWAP_FINISHED = Events.SetChanged,
        PLAYER_EQUIPMENT_CHANGED = Events.EquipmentChanged,
    };

    -- These are only registered if the plugin is active as they control how what is called
    -- from LunaEclipse Rotation Helper
    addon.System.Callbacks.RegisterAddonCallbacks(moduleEvents);
    addon.System.Events.RegisterEvents(tableEvents, true);
end