-- Pulls back the Addon-Local Variables and store them locally.
local _, addonTable = ...;
local addon = _G["LunaEclipse_Rotation"];

-- Plugin information variables.
local pluginName = "Equipment Manager";

-- Check whether the plugin is supposed to be active
if addon and addon.System.Plugins.IsActive(pluginName) then
    -- Store local copies of Blizzard API and other addon global functions and variables.
    local ITEM_INVENTORY_LOCATION_BAGS, ITEM_INVENTORY_LOCATION_BANK, ITEM_INVENTORY_LOCATION_PLAYER = ITEM_INVENTORY_LOCATION_BAGS, ITEM_INVENTORY_LOCATION_BANK, ITEM_INVENTORY_LOCATION_PLAYER;
    local bit, pairs, string = bit, pairs, string;

    -- List of inventory slot names and index.
    local INVENTORY_SLOTS = {
        [1] = "HeadSlot",
        [2] = "NeckSlot",
        [3] = "ShoulderSlot",
        [4] = "ShirtSlot",
        [5] = "ChestSlot",
        [6] = "WaistSlot",
        [7] = "LegsSlot",
        [8] = "FeetSlot",
        [9] = "WristSlot",
        [10] = "HandsSlot",
        [11] = "Finger0Slot",
        [12] = "Finger1Slot",
        [13] = "Trinket0Slot",
        [14] = "Trinket1Slot",
        [15] = "BackSlot",
        [16] = "MainHandSlot",
        [17] = "SecondaryHandSlot",
        [19] = "TabardSlot"
    };

    -- Variables for plugin.
    local overlayFrames = {};
    local textureInBank = { red = 0, green = 1, blue = 1, alpha = 0.4 };
    local textureInBags = { red = 0, green = 1, blue = 0, alpha = 0.4 };
    local textureNotUsed = { red = 1, green = 1, blue = 0, alpha = 0.4 };
    local textureNotAvailable = { red = 1, green = 0, blue = 0, alpha = 0.4 };

    -- Tables to hold private functions.
    local Overlay = {};

    --------------------- Private Module Functions ---------------
    Overlay = {
        ChooseColor = function(slotID, itemLocation, currentItemID, itemID)
            if slotID and itemLocation then
                local player = bit.band(itemLocation, ITEM_INVENTORY_LOCATION_PLAYER) ~= 0;
                local bank = bit.band(itemLocation, ITEM_INVENTORY_LOCATION_BANK) ~= 0;
                local bags = bit.band(itemLocation, ITEM_INVENTORY_LOCATION_BAGS) ~= 0;

                if player and not bags then
                    Overlay.Tint(slotID);
                elseif bags and not bank then
                    Overlay.Tint(slotID, textureInBags);
                elseif bank then
                    Overlay.Tint(slotID, textureInBank);
                else
                    Overlay.Tint(slotID, textureNotAvailable);
                end
            elseif slotID and not itemLocation then
                if currentItemID == itemID then
                    Overlay.Tint(slotID);
                else
                    Overlay.Tint(slotID, textureNotUsed);
                end
            end
        end,

        Hide = function()
            for slotID in pairs(INVENTORY_SLOTS) do
                Overlay.Tint(slotID);
            end
        end,

        Show = function()
            local Tables = addon.Core.Tables;
            local equipManager = addonTable.equipManager;

            if PaperDollEquipmentManagerPane and PaperDollEquipmentManagerPane:IsShown() and Tables.GetValue(equipManager, "useOverlay", 1) == 1 then
                local selectedSetID = PaperDollEquipmentManagerPane.selectedSetID;

                if selectedSetID then
                    local setInfo = addon.Features.EquipmentManager.Info(selectedSetID);

                    if not Tables.IsEmpty(setInfo) then
                        for slotID in pairs(INVENTORY_SLOTS) do
                            local itemLocation = Tables.GetValue(setInfo.Locations, slotID);
                            local currentItemID = addon.API.Blizzard.GetInventoryItemID("player", slotID) or 0;
                            local setItemID = Tables.GetValue(setInfo.ItemIDs, slotID, 0);

                            Overlay.ChooseColor(slotID, itemLocation, currentItemID, setItemID);
                        end
                    end
                end
            end
        end,

        Tint = function(slotID, colorTable)
            if slotID then
                local slotName = INVENTORY_SLOTS[slotID];
                local slotFrame = _G[string.format("Character%s", slotName)];
                local overlayFrame = overlayFrames[slotID];

                if slotFrame and not overlayFrame then
                    overlayFrame = slotFrame:CreateTexture(nil, "OVERLAY");
                    overlayFrame:SetAllPoints();

                    overlayFrames[slotID] = overlayFrame;
                end

                if colorTable then
                    overlayFrame:SetColorTexture(colorTable.red, colorTable.green, colorTable.blue, colorTable.alpha);
                else
                    overlayFrame:SetColorTexture(0, 0, 0, 0);
                end
            end
        end,
    };

    ----------------------- Plugin Code Operation ----------------
    local moduleEvents = {
        playerLogin = function()
            addon:SecureHook(nil, "GearSetButton_OnClick", Overlay.Show);
            addon:HookScript(PaperDollEquipmentManagerPane, "OnShow", Overlay.Show);
            addon:HookScript(PaperDollEquipmentManagerPane, "OnHide", Overlay.Hide);
        end,

        playerLogout = function()
            addon:Unhook(PaperDollEquipmentManagerPane, "OnShow");
            addon:Unhook(PaperDollEquipmentManagerPane, "OnHide");
            addon:Unhook(nil, "GearSetButton_OnClick");

            Overlay.Hide();
        end,
    };

    local tableEvents = {
        BAG_UPDATE = Overlay.Show,
        BANKFRAME_OPENED = Overlay.Show,
        BANKFRAME_CLOSED = Overlay.Show,
        EQUIPMENT_SETS_CHANGED = Overlay.Show,
        PLAYER_EQUIPMENT_CHANGED = Overlay.Show,
        PLAYERBANKSLOTS_CHANGED = Overlay.Show,
    };

    -- These are only registered if the plugin is active as they control how what is called
    -- from LunaEclipse Rotation Helper
    addon.System.Callbacks.RegisterAddonCallbacks(moduleEvents);
    addon.System.Events.RegisterEvents(tableEvents, true);
end