-- Pulls back the Addon-Local Variables and store them locally.
local _, addonTable = ...;
local addon = _G["LunaEclipse_Rotation"];

-- Plugin information variables.
local pluginName = "Interrupt Manager";
local pluginDescription = "This plugin suggests interrupts on the interrupt icon.\n\nInterrupts can be shown even when you don't have a rotation loaded if desired!";

-- Check for LunaEclipse Rotation Helper being loaded first so it doesn't register the plugin if its not needed.
if addon then
    -- Store local copies of Blizzard API and other addon global functions and variables
    local format = format;

    local groupText = { "solo", "in a party", "in a raid" };
    local groupType = { "Solo", "Party", "Raid" };
    local interruptGroupType = 1;

    local optionsTable = {
        type = "group",
        name = pluginName,
        get = function(info)
            local returnValue = 0;

            if info then
                local Tables = addon.Core.Tables;
                local profile = addonTable.interruptManager or {};
                local keyName = info[#info];

                returnValue = Tables.GetValue(profile, keyName);

                if returnValue == nil then
                    returnValue = 0;
                    profile[keyName] = returnValue;

                    addonTable.interruptManager = profile;
                end

                if info.option.type == "toggle" then
                    returnValue = addon.Options.Core.ConvertToUI(returnValue, info.option.tristate);
                end
            end

            return returnValue;
        end,
        set = function(info, value)
            if info then
                local profile = addonTable.interruptManager or {};
                local keyName = info[#info];

                if info.option.type == "toggle" then
                    value = addon.Options.Core.ConvertToValue(value, info.option.tristate);
                end

                profile[keyName] = value;

                -- Put the settings into the database
                addonTable.interruptManager = profile;
            end
        end,
        childGroups = "tree",
        args = {
            requireRotation = {
                type = "toggle",
                name = "Only suggest interrupts when Rotation is loaded.",
                desc = "This will show interrupts if you are currently using a rotation provided by the addon.\n\nIf this is unchecked then interrupts will be suggested even for unsupported rotations.",
                width = "full",
                order = 10,
            },
            useCC = {
                type = "toggle",
                name = "Use crowd control abilities to interrupt.",
                desc = "This will recommend crowd control abilities to interrupt spell casting.\n\nWarning: Using crowd control abilities to interrupt spell casting is unreliable, and the target usually begins casting immediately after the crowd control effect has finished.\n\nThis is defaulted to unchecked.",
                width = "full",
                order = 20,
            },
            useStun = {
                type = "toggle",
                name = "Use stuns to interrupt, when the target is stunnable.",
                desc = "This will recommend stuns to interrupt spell casting if the target is stunnable.\n\nThis can be mildly useful if your interrupt is on cooldown, but as not all enemies are stunnable this usefulness of this is limited.\n\nIn raids this is really only useful against trash.",
                width = "full",
                order = 30,
            },
            InterruptOptions = {
                type = "group",
                name = addon.Options.Core.CreaterHeader("Interrupt Limitation"),
                get = function(info)
                    local coreTables = addon.Core.Tables;
                    local returnValue = true;

                    if info then
                        local sectionName = info[#info - 1];
                        local keyName = format("%s%s", info[#info], groupType[interruptGroupType]);
                        local profile = coreTables.GetValue(addonTable.interruptManager, sectionName, {});

                        returnValue = coreTables.GetValue(profile, keyName);

                        if info.option.type == "toggle" then
                            returnValue = addon.Options.Core.ConvertToUI(returnValue, info.option.tristate);
                        end
                    end

                    return returnValue;
                end,
                set = function(info, value)
                    if info then
                        local coreTables = addon.Core.Tables;

                        if info.option.type == "toggle" then
                            value = addon.Options.Core.ConvertToValue(value, info.option.tristate);
                        end

                        local sectionName = info[#info - 1];
                        local keyName = info[#info];
                        local profile = coreTables.GetValue(addonTable.interruptManager, sectionName, {});

                        if keyName == "noLimits" then
                            if value == 1 then
                                profile[format("%s%s", "limitPet", groupType[interruptGroupType])] = 0;
                                profile[format("%s%s", "limitPlayer", groupType[interruptGroupType])] = 0;
                                profile[format("%s%s", "limitFriendly", groupType[interruptGroupType])] = 0;
                            end
                        else
                            if value == 1 then
                                profile[format("%s%s", "noLimits", groupType[interruptGroupType])] = 0;
                            end
                        end

                        keyName = format("%s%s", info[#info], groupType[interruptGroupType]);
                        profile[keyName] = value;

                        -- Put the settings into the database
                        addonTable.interruptManager[sectionName] = profile;
                    end
                end,
                order = 40,
                inline = true,
                args = {
                    grouptType = {
                        type = "select",
                        name = "Group Type",
                        desc = "Select which type of group you want to limit interrupt suggestions for.\n\nIf you want different limits for each group type you must select each group type and then the settings, then change to the next group type and continue.",
                        get = function(info)
                            return interruptGroupType;
                        end,
                        set = function(info, value)
                            interruptGroupType = value;
                        end,
                        values = {
                            [1] = "Solo",
                            [2] = "Party",
                            [3] = "Raid",
                        },
                        width = "full",
                        order = 10,
                    },
                    noLimits = {
                        type = "toggle",
                        name = "No limits.",
                        desc = "Selecting this will suggest interrupts for all interruptible spells if you have an available interrupt.",
                        width = "full",
                        order = 20,
                    },
                    limitPlayer = {
                        type = "toggle",
                        name = "Only interrupt casts targetting you.",
                        desc = "Selecting this will only suggest interrupts for spells cast against you if you have an available interrupt.\n\nIf you have other options checked these will also be processed.\n\nSelecting 'No limits' overrides this selection!",
                        width = "full",
                        order = 30,
                    },
                    limitPet = {
                        type = "toggle",
                        name = "Only interrupt casts targetting your pet.",
                        desc = "Selecting this will only suggest interrupts for spells cast against your pet if you have an available interrupt.\n\nIf you have other options checked these will also be processed.\n\nSelecting 'No limits' overrides this selection!",
                        width = "full",
                        order = 40,
                    },
                    limitFriendly = {
                        type = "toggle",
                        name = "Only interrupt casts at units friendly to the caster.",
                        desc = "Selecting this will only suggest interrupts for spells cast against targets that are friendly to the caster.  This is useful if you only want to interrupt heals or buffs.\n\nIf you have other options checked these will also be processed.\n\nSelecting 'No limits' overrides this selection!",
                        width = "full",
                        order = 50,
                    },
                    description = {
                        type = "description",
                        name = format("\nIf you select 'No limits' all other options will be unchecked as there will be no limitations on suggesting interrupts.\n\nIf you select any of the other three options then 'No Limits' will be unchecked as interrupts will now be limited.\n\nIf you do not wish interrupts to be suggested at all, uncheck all the options.\n\nCurrently you are selecting the options for when you are playing %s!", groupText[interruptGroupType]),
                        width = "full",
                        order = 60,
                    },
                },
            },
        },
    };

    -- The plugin still has to be registered even if its not loaded into memory, otherwise there is no checkbox to re-enable it.
    addon.System.Plugins.RegisterPlugin(pluginName, pluginDescription, optionsTable);
end