-- Pulls back the Addon-Local Variables and store them locally.
local _, addonTable = ...;
local addon = _G["LunaEclipse_Rotation"];

-- Plugin information variables.
local pluginName = "Macro Manager";

-- Check whether the plugin is supposed to be active
if addon and addon.System.Plugins.IsActive(pluginName) then
    -- Store local copies of Blizzard API and other addon global functions and variables.
    local pairs, table = pairs, table;

    -- Local variables for plugin operation.
    local changingSpec = false;

    -- Tables to hold private functions.
    local Buttons, Events, Macros, Profiles = {}, {}, {}, {};

    --------------------- Private Module Functions ---------------
    Buttons = {
        Clear = function(buttonID)
            addon.API.Blizzard.ClearCursor();
            addon.API.Blizzard.PickupAction(buttonID);
            addon.API.Blizzard.ClearCursor();
        end,

        Count = function()
            local returnValue = 0;

            for counter = 1, 120 do
                local actionType, actionID = addon.API.Blizzard.GetActionInfo(counter);

                if (actionType and actionType == "macro") and (actionID and actionID >= 121 and actionID <= 138) then
                    returnValue = returnValue + 1;
                end
            end

            return returnValue;
        end,

        Restore = function(buttonID, macroName)
            local macroID = addon.API.Blizzard.GetMacroIndexByName(macroName);

            addon.API.Blizzard.ClearCursor();
            addon.API.Blizzard.PickupMacro(macroID);
            addon.API.Blizzard.PlaceAction(buttonID);
            addon.API.Blizzard.ClearCursor();
        end,
    };

    Macros = {
        ClearMacros = function()
            for counter = 138, 121, -1 do
                addon.API.Blizzard.DeleteMacro(counter);
            end
        end,

        Current = function()
            local returnValue = {};

            for counter = 121, 138 do
                local macroInfo = Macros.GetMacro(counter);

                if macroInfo then
                    returnValue[macroInfo.Name] = {
                        Body = macroInfo.Body,
                        ButtonIDs = macroInfo.ButtonIDs,
                    };
                end
            end

            return returnValue;
        end,

        GetButtonIDs = function(macroID)
            local returnValue = {};

            for counter = 1, 120 do
                local actionType, actionID = addon.API.Blizzard.GetActionInfo(counter);

                if (actionType and actionType == "macro") and ((macroID ~= nil and actionID and actionID == macroID) or (macroID == nil and actionID and actionID >= 121 and actionID <= 138)) then
                    table.insert(returnValue, counter);
                end
            end

            return returnValue;
        end,

        GetMacro = function(macroID)
            local returnValue;
            local macroButtonIDs = Macros.GetButtonIDs(macroID);
            local macroName, _, macroBody = addon.API.Blizzard.GetMacroInfo(macroID);

            if macroName and macroBody then
                returnValue = {
                    Name = macroName,
                    Body = macroBody,
                    ButtonIDs = macroButtonIDs,
                };
            end

            return returnValue;
        end,
    };

    Profiles = {
        CreateButtons = function(profile)
            local Tables = addon.Core.Tables;

            if not Tables.IsEmpty(profile) then
                for macroName, macroData in pairs(profile) do
                    if not Tables.IsEmpty(macroData.ButtonIDs) then
                        for _, buttonID in pairs(macroData.ButtonIDs) do
                            Buttons.Clear(buttonID);
                            Buttons.Restore(buttonID, macroName);
                        end
                    end
                end
            end
        end,

        CreateMacros = function(profile)
            local Tables = addon.Core.Tables;

            if not Tables.IsEmpty(profile) then
                for macroName, macroData in pairs(profile) do
                    local macroBody = macroData.Body;
                    local macroTexture = "INV_MISC_QUESTIONMARK";

                    addon.API.Blizzard.CreateMacro(macroName, macroTexture, macroBody, 1);
                end
            end
        end,

        Exists = function(specInfo)
            local Tables = addon.Core.Tables;
            local addonSettings = addonTable.addonSettings;
            local returnValue = false;

            if specInfo then
                local specName = Tables.GetValue(specInfo, "Name", "");
                local profile = Tables.GetValue(addonTable.macroProfiles, specName);

                if Tables.GetValue(addonSettings, "allowZeroMacros", 0) == 1 then
                    returnValue = profile ~= nil or false;
                else
                    returnValue	= profile ~= nil and not Tables.IsEmpty(profile) or false;
                end
            end

            return returnValue;
        end,

        Load = function(specInfo)
            local Tables = addon.Core.Tables;

            if not Tables.IsEmpty(specInfo) then
                local specName = Tables.GetValue(specInfo, "Name", "");
                local savedProfile = Tables.GetValue(addonTable.macroProfiles, specName, {});
                local currentMacros = Macros.Current();

                if not Tables.Equate(savedProfile, currentMacros) then
                    Macros.ClearMacros();
                    Profiles.CreateMacros(savedProfile);
                    Profiles.CreateButtons(savedProfile);

                    addon.System.Timers.TimerAdd(Profiles.Verify, 0.25, true, specInfo);
                end
            end
        end,

        Save = function(specInfo, forceSave)
            local Tables = addon.Core.Tables;

            if not Tables.IsEmpty(specInfo) then
                local specName = Tables.GetValue(specInfo, "Name", "");
                local savedProfile = Tables.GetValue(addonTable.macroProfiles, specName, {});
                local currentMacros = Macros.Current();

                if forceSave or (not Tables.Equate(savedProfile, currentMacros) and Buttons.Count() > 0) then
                    addonTable.macroProfiles[specName] = currentMacros;
                    addon.System.Callbacks.Callback("updateMacros");
                end
            end
        end,

        Update = function(profiles)
            local Tables = addon.Core.Tables;

            local currentProfile = Tables.Duplicate(profiles);
            local returnValue = {};

            if not Tables.IsEmpty(currentProfile) then
                for specName, macroProfile in pairs(currentProfile) do
                    returnValue[specName] = {};

                    if not Tables.IsEmpty(macroProfile) then
                        for macroName, macroData in pairs(macroProfile) do
                            local newMacroData = {
                                Body = macroData.Body,
                                ButtonIDs = macroData.ButtonIDs,
                            };

                            returnValue[specName][macroName] = newMacroData;
                        end
                    end
                end
            end

            return returnValue;
        end,

        Verify = function(specInfo)
            local Tables = addon.Core.Tables;

            local currentProfile = Macros.Current();
            local savedProfile = Tables.GetValue(addonTable.macroProfiles, specInfo.Name or "", {});

            local different = false;

            if not Tables.IsEmpty(savedProfile) then
                for macroName, macroData in pairs(savedProfile) do
                    local currentMacro = currentProfile[macroName];

                    if currentMacro and Tables.GetValue(macroData, "Body", "") == Tables.GetValue(currentMacro, "Body", "")
                                    and not Tables.IsEmpty(macroData.ButtonIDs) then
                        for _, buttonID in pairs(macroData.ButtonIDs) do
                            local actionType, actionID = addon.API.Blizzard.GetActionInfo(buttonID);
                            local macroID = addon.API.Blizzard.GetMacroIndexByName(macroName);

                            if not actionType == "macro" or actionID ~= macroID then
                                Buttons.Clear(buttonID);
                                Buttons.Restore(buttonID, macroName);

                                different = true;
                            end
                        end
                    elseif not currentMacro then
                        different = true;
                    end
                end
            end

            if not different then
                addon.System.Timers.TimerRemove(Profiles.Verify);

                if not Tables.IsEmpty(savedProfile) then
                    addon.System.Callbacks.Callback("updateMacros");
                end
            end
        end,
    };

    ------------------------ WOW API Events ----------------------
    Events = {
        MacroChanged = function()
            local Tables = addon.Core.Tables;
            local apiGeneral = addon.API.General;

            local specInfo = apiGeneral.GetSpecializationInfo();

            if not changingSpec and not Tables.IsEmpty(specInfo) then
                addon.Features.MacroManager.Save(specInfo);
            end
        end,
    };

    ---------------------- Public Functions ----------------------
    addon.Features.MacroManager = {
        Load = function(specInfo)
            addon.System.Events.CombatEventDelay(Profiles.Load, specInfo);
        end,

        Save = function(specInfo, forceSave)
            Profiles.Save(specInfo, forceSave);
        end,

        Update = function(profiles)
            return Profiles.Update(profiles);
        end,
    };

    ----------------------- Plugin Code Operation ----------------
    local moduleEvents = {
        playerLogin = function(specInfo)
            addon.System.Timers.TimerAdd(Profiles.Verify, 0.25, true, specInfo);
        end,

        changeSpecFailed = function()
            changingSpec = false;
        end,

        changeSpecFinish = function(specInfo)
            if specInfo then
                if not Profiles.Exists(specInfo) then
                    addon.Features.MacroManager.Save(specInfo, true);
                else
                    addon.Features.MacroManager.Load(specInfo);
                end

                changingSpec = false;
            end
        end,

        changeSpecStart = function(specInfo)
            changingSpec = true;
            addon.Features.MacroManager.Save(specInfo);
        end,
    };

    local tableEvents = {
        ACTIONBAR_SLOT_CHANGED = Events.MacroChanged,
        UPDATE_MACROS = Events.MacroChanged,
    };

    -- These are only registered if the plugin is active as they control how what is called
    -- from LunaEclipse Rotation Helper
    addon.System.Callbacks.RegisterAddonCallbacks(moduleEvents);
    addon.System.Events.RegisterEvents(tableEvents, true);
end