-- Pulls back the Addon-Local Variables and store them locally.
local _, addonTable = ...;
local addon = _G["LunaEclipse_Rotation"];

-- Plugin information variables.
local pluginName = "Minimap Icon/LDB Display";

-- Required libraries
local baseDataBroker = LibStub("LibDataBroker-1.1");
local addonIcon = LibStub("LibDBIcon-1.0");
local addonTooltip = LibStub("LibQTip-1.0");

-- Check whether the plugin is supposed to be active and that the required Weak Auras is loaded.
if addon and addon.System.Plugins.IsActive(pluginName) and baseDataBroker and addonIcon and addonTooltip then
    -- Store local copies of Blizzard API and other addon global functions and variables.
    local pairs, string, tostring = pairs, string, tostring;
    local HIGHLIGHT_FONT_COLOR_CODE, LOOT_SPECIALIZATION_DEFAULT, SELECT_LOOT_SPECIALIZATION, SPECIALIZATION = HIGHLIGHT_FONT_COLOR_CODE, LOOT_SPECIALIZATION_DEFAULT, SELECT_LOOT_SPECIALIZATION, SPECIALIZATION;

    -- Local variables for plugin operation.
    local dataTextName = "LunaEclipse: Rotation Helper";
    local menuName = "LERH_Minimap_Menu";
    local minimapMenu = addon.API.Blizzard.CreateFrame("Frame", menuName, UIParent, "UIDropDownMenuTemplate");

    local dataBroker, iconTooltip, reloadRequired;

    local LDBDefaults = {
        icon = "Interface\\ICONS\\spell_nature_bloodlust",
        hide = false,
    };

    local clickTexture = {
        texture = "Interface\\TUTORIALFRAME\\UI-TUTORIAL-FRAME",
        height = "16", width = "12",
        xOffset = "0", yOffset = "-1",
        xDimension = "512",	yDimension = "512",
        leftClick = {
            left = "9",	right = "69", top = "228", bottom = "308",
        },
        rightClick = {
            left = "9",	right = "69", top = "330", bottom = "410",
        },
    };

    local leftButtonTextureString = string.format("|T%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s|t", clickTexture.texture, clickTexture.height, clickTexture.width, clickTexture.xOffset, clickTexture.yOffset, clickTexture.xDimension, clickTexture.yDimension, clickTexture.leftClick.left, clickTexture.leftClick.right, clickTexture.leftClick.top, clickTexture.leftClick.bottom);
    local rightButtonTextureString = string.format("|T%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s|t", clickTexture.texture, clickTexture.height, clickTexture.width, clickTexture.xOffset, clickTexture.yOffset, clickTexture.xDimension, clickTexture.yDimension, clickTexture.rightClick.left, clickTexture.rightClick.right, clickTexture.rightClick.top, clickTexture.rightClick.bottom);

    -- Tables to hold private functions.
    local Icon, Mouse, Tooltip = {}, {}, {};

    --------------------- Popup Dialog Boxes ---------------------
    StaticPopupDialogs["LERH_CONFIRM_RELOAD_UI"] = {
        text = string.join("", "Please confirm you wish to Reload UI.", "\n\n", "This ensures that your profile currently loaded in memory is written to your hard disk!", "\n\n"),
        button1	= "Yes",
        button2	= "No",
        OnAccept = function(self)
            addon.API.Blizzard.ReloadUI();
        end,
        timeout	= 10,
        whileDead = true,
        hideOnEscape = true,
    };

    StaticPopupDialogs["LERH_TOGGLE_MINIMAP"] = {
        text = string.format("%s%s%s%s", "A Reload UI is required for the icon to become visible when using MiniMap icon manager addons.", "\n\n", "Do you wish to reload now?", "\n\n"),
        button1	= "Yes",
        button2	= "No",
        OnAccept = function()
            addon.API.Blizzard.ReloadUI();
        end,
        timeout	= 15,
        whileDead = true,
        hideOnEscape = true,
    };

    --------------------- Private Module Functions ---------------
    Icon = {
        Create = function(specInfo)
            local Tables = addon.Core.Tables;

            local minimapSettings = addonTable.minimapSettings;
            local LDBIconData = {
                type = "data source",
                OnClick = Icon.OnClick,
                OnLeave = Icon.OnLeave,
                OnEnter = Icon.OnEnter,
            };

            LDBIconData = Tables.Merge(LDBIconData, LDBDefaults);
            LDBIconData = Tables.Merge(LDBIconData, minimapSettings);

            reloadRequired = not minimapSettings.hide;
            dataBroker = baseDataBroker:NewDataObject(dataTextName, LDBIconData);

            addonIcon:Register(dataTextName, dataBroker, minimapSettings);

            Icon.UpdateLDB(specInfo);
        end,

        OnClick = function(self, button)
            local Tables = addon.Core.Tables;
            local minimapSettings = addonTable.minimapSettings;

            local changeLootSpec = Tables.GetValue(minimapSettings, "changeLootSpec", 1) == 1;
            local changeSpec = Tables.GetValue(minimapSettings, "changeSpec", 1) == 1;
            local reloadUI = Tables.GetValue(minimapSettings, "reloadUI", 1) == 1;
            local toggleIcon = Tables.GetValue(minimapSettings, "toggleIcon", 1) == 1;
            local useRotations = Tables.GetValue(minimapSettings, "useRotations", 1) == 1;
            local useTalentBuilds = addon.System.Plugins.IsActive("Talent Builds") and Tables.GetValue(minimapSettings, "useTalentBuilds", 1) == 1;

            if button == "LeftButton" then
                if addon.API.Blizzard.IsControlKeyDown() and changeSpec then
                    Mouse.Specialization();
                elseif addon.API.Blizzard.IsShiftKeyDown() and toggleIcon then
                    addon.Features.MinimapIcon.Toggle();
                elseif not addon.API.Blizzard.IsModifierKeyDown() and useRotations then
                    local apiGeneral = addon.API.General;

                    local specInfo = apiGeneral.GetSpecializationInfo();

                    if not Tables.IsEmpty(specInfo) then
                        local numRotations = addon.Rotation.GetRegisteredRotations(specInfo.ID);

                        if numRotations >= 1 then
                            Mouse.Rotations();
                        end
                    end
                end
            elseif button == "RightButton" then
                if addon.API.Blizzard.IsControlKeyDown() and changeLootSpec then
                    Mouse.LootSpecialization();
                elseif addon.API.Blizzard.IsShiftKeyDown() and reloadUI then
                    addon.API.Blizzard.StaticPopup_Show("LERH_CONFIRM_RELOAD_UI");
                elseif not addon.API.Blizzard.IsModifierKeyDown() and useTalentBuilds then
                    local TalentBuilds = addon.Features.TalentBuilds;

                    local buildsAvailable = useTalentBuilds and not Tables.IsEmpty(TalentBuilds.Available());
                    local changesAllowed = useTalentBuilds and TalentBuilds.ChangesAllowed();

                    if buildsAvailable and changesAllowed then
                        Mouse.Builds();
                    end
                end
            end
        end,

        OnEnter = function(self)
            iconTooltip = addonTooltip:Acquire("LERH_Minimap_ToolTip", 1, "LEFT");
            iconTooltip:SmartAnchorTo(self);
            Tooltip.Setup();
        end,

        OnLeave = function()
            addonTooltip:Release(iconTooltip);
            iconTooltip = nil;
        end,

        UpdateLDB = function(specInfo)
            if specInfo and dataBroker then
                local Tables = addon.Core.Tables;
                local minimapSettings = addonTable.minimapSettings;

                local useTalentBuilds = addon.System.Plugins.IsActive("Talent Builds") and Tables.GetValue(minimapSettings, "useTalentBuilds", 1) == 1;
                local talentText = Tables.GetValue(specInfo, "IconID") ~= nil and string.format("|T%s:14:14:0:0:64:64:4:60:4:60|t", specInfo.IconID) or "";

                if useTalentBuilds then
                    local TalentBuilds = addon.Features.TalentBuilds;
                    local buildInfo = TalentBuilds.ActiveBuild();

                    dataBroker.text = string.format("%s %s: %s", talentText, "Build", buildInfo.Name) or LDBDefaults.text;
                else
                    dataBroker.text = string.format("%s %s: %s", talentText, "Spec", specInfo.Name) or LDBDefaults.text;
                end

                dataBroker.icon = Tables.GetValue(specInfo, "IconID", LDBDefaults.icon);

                for key, value in pairs(minimapSettings) do
                    dataBroker[key] = value;
                end

                addonIcon:Refresh(dataTextName, minimapSettings);
                Tooltip.Setup();
            end
        end,

        UpdateToolTip = function()
            Tooltip.Setup();
        end,
    };

    Mouse = {
       Actions = function()
            local Tables = addon.Core.Tables;
            local minimapSettings = addonTable.minimapSettings;
            local returnValue = false;

            if not addon.API.Blizzard.IsModifierKeyDown() then
                if Tables.GetValue(minimapSettings, "useRotations", 1) == 1 then
                    local apiGeneral = addon.API.General;
                    local specInfo = apiGeneral.GetSpecializationInfo();

                    if not Tables.IsEmpty(specInfo) then
                        local numRotations = addon.Rotation.GetRegisteredRotations(specInfo.ID);

                        returnValue = numRotations >= 1;
                    end
                end

                if not returnValue and addon.System.Plugins.IsActive("Talent Builds") and Tables.GetValue(minimapSettings, "useTalentBuilds", 1) == 1 then
                    local TalentBuilds = addon.Features.TalentBuilds;

                    returnValue = not Tables.IsEmpty(TalentBuilds.Available()) and TalentBuilds.ChangesAllowed();
                end
            elseif addon.API.Blizzard.IsControlKeyDown() then
                returnValue = Tables.GetValue(minimapSettings, "changeSpec", 1) == 1 or Tables.GetValue(minimapSettings, "changeLootSpec", 1) == 1;
            elseif addon.API.Blizzard.IsShiftKeyDown() then
                returnValue = Tables.GetValue(minimapSettings, "toggleIcon", 1) == 1 or Tables.GetValue(minimapSettings, "reloadUI", 1) == 1;
            end

            return returnValue;
        end,

        Builds = function()
            local Tables = addon.Core.Tables;
            local TalentBuilds = addon.Features.TalentBuilds;

            local availableBuilds = TalentBuilds.Available();

            if not Tables.IsEmpty(availableBuilds) then
                local buildInfo = TalentBuilds.ActiveBuild();
                local counter = 2;

                if iconTooltip then
                    iconTooltip:Hide();
                end

                local availableBuildsList = {
                    { text = "Available Builds", isTitle = true, notCheckable = true },
                };

                for buildName in pairs(availableBuilds) do
                    local isActive = buildInfo.Name == buildName;

                    availableBuildsList[counter] = availableBuildsList[counter] or {};

                    availableBuildsList[counter].text = string.format("%s%s%s%s", isActive and addonTable.COLORS.TAGS.ACTIVE or addonTable.COLORS.TAGS.INACTIVE, buildName, isActive and " (Active)" or "", addonTable.TAGS.END);
                    availableBuildsList[counter].func = function()
                        TalentBuilds.Apply(buildName);
                    end;

                    availableBuildsList[counter].notCheckable = true;

                    counter = counter + 1;
                end

                addon.API.Blizzard.EasyMenu(availableBuildsList, minimapMenu, "cursor", -15, -7, "MENU", 2);
            end
        end,

        LootSpecialization = function()
            local Tables = addon.Core.Tables;
            local apiGeneral = addon.API.General;

            local specInfo = apiGeneral.GetSpecializationInfo();

            if not Tables.IsEmpty(specInfo) then
                local numSpecializations = Tables.GetValue(specInfo, "TotalSpecializations", 0);
                local lootSpecialization = addon.API.Blizzard.GetLootSpecialization();

                if iconTooltip then
                    iconTooltip:Hide();
                end

                local availableLootSpecList = {
                    { text = SELECT_LOOT_SPECIALIZATION, isTitle = true, notCheckable = true },
                };

                availableLootSpecList[2] = {
                    text = string.format("%s%s%s", lootSpecialization == 0 and addonTable.COLORS.TAGS.ACTIVE or addonTable.COLORS.TAGS.INACTIVE, string.format(LOOT_SPECIALIZATION_DEFAULT, Tables.GetValue(specInfo, "Name") or "None"), addonTable.TAGS.END),
                    func = function()
                        addon.API.Blizzard.SetLootSpecialization(0);
                    end,
                    notCheckable = true,
                };

                for counter = 1, numSpecializations do
                    local newSpecInfo = apiGeneral.GetSpecializationInfo(false, counter);

                    if not Tables.IsEmpty(newSpecInfo) then
                        availableLootSpecList[counter + 2] = availableLootSpecList[counter + 2] or {};

                        local isActive = lootSpecialization == Tables.GetValue(newSpecInfo, "ID");

                        availableLootSpecList[counter + 2].text = string.format("%s%s%s", isActive and addonTable.COLORS.TAGS.ACTIVE or addonTable.COLORS.TAGS.INACTIVE, Tables.GetValue(newSpecInfo, "Name"), addonTable.TAGS.END);
                        availableLootSpecList[counter + 2].func = function()
                            addon.API.Blizzard.SetLootSpecialization(counter);
                        end;

                        availableLootSpecList[counter + 2].notCheckable = true;
                    else
                        availableLootSpecList[counter + 2] = nil;
                    end
                end

                addon.API.Blizzard.EasyMenu(availableLootSpecList, minimapMenu, "cursor", -15, -7, "MENU", 2);
            end
        end,

        Rotations = function()
            local Rotation = addon.Rotation;
            local Tables = addon.Core.Tables;
            local apiGeneral = addon.API.General;

            local specInfo = apiGeneral.GetSpecializationInfo();

            if not Tables.IsEmpty(specInfo) then
                local _, rotationList = Rotation.GetRegisteredRotations(Tables.GetValue(specInfo, "ID"));

                if not Tables.IsEmpty(rotationList) then
                    local activeRotation = addon.Rotation.GetActiveRotation();
                    local counter = 3;

                    if iconTooltip then
                        iconTooltip:Hide();
                    end

                    local availableRotationList = {
                        { text = "Available Rotations", isTitle = true, notCheckable = true },
                    };

                    local isActive = activeRotation == "__none";

                    availableRotationList[2] = availableRotationList[2] or {};
                    availableRotationList[2].text = string.format("%s%s%s%s", isActive and addonTable.COLORS.TAGS.ACTIVE or addonTable.COLORS.TAGS.INACTIVE, "No Rotation Selected", isActive and " (Active)" or "", addonTable.TAGS.END);
                    availableRotationList[2].func = function()
                        Rotation.SetActiveRotation("__none");
                    end;

                    availableRotationList[2].notCheckable = true;

                     for rotationName, rotation in pairs(rotationList) do
                        isActive = activeRotation == rotationName;
                        availableRotationList[counter] = availableRotationList[counter] or {};

                        availableRotationList[counter].text = string.format("%s%s%s%s", isActive and addonTable.COLORS.TAGS.ACTIVE or addonTable.COLORS.TAGS.INACTIVE, rotation["description"], isActive and " (Active)" or "", addonTable.TAGS.END);
                        availableRotationList[counter].func = function()
                            Rotation.SetActiveRotation(rotationName);
                        end;

                        availableRotationList[counter].notCheckable = true;
                        counter = counter + 1;
                    end

                    addon.API.Blizzard.EasyMenu(availableRotationList, minimapMenu, "cursor", -15, -7, "MENU", 2);
                end
            end
        end,

        Specialization = function()
            local Tables = addon.Core.Tables;
            local apiGeneral = addon.API.General;

            local specInfo = apiGeneral.GetSpecializationInfo();

            if not Tables.IsEmpty(specInfo) then
                local specIndex = Tables.GetValue(specInfo, "Index");
                local numSpecializations = Tables.GetValue(specInfo, "TotalSpecializations", 0);

                if iconTooltip then
                    iconTooltip:Hide();
                end

                local availableSpecList = {
                    { text = SPECIALIZATION, isTitle = true, notCheckable = true },
                };

                for counter = 1, numSpecializations do
                    local newSpecInfo = apiGeneral.GetSpecializationInfo(false, counter);

                    if not Tables.IsEmpty(newSpecInfo) then
                        availableSpecList[counter + 1] = availableSpecList[counter + 1] or {};

                        local isActive = specIndex == counter;

                        availableSpecList[counter + 1].text = string.format("%s%s%s", (isActive and addonTable.COLORS.TAGS.ACTIVE) or addonTable.COLORS.TAGS.INACTIVE, Tables.GetValue(newSpecInfo, "Name"), addonTable.TAGS.END);
                        availableSpecList[counter + 1].func = function()
                            addon.API.Blizzard.SetSpecialization(counter);
                        end;

                        availableSpecList[counter + 1].notCheckable = true;
                    else
                        availableSpecList[counter + 1] = nil;
                    end
                end

                addon.API.Blizzard.EasyMenu(availableSpecList, minimapMenu, "cursor", -15, -7, "MENU", 2);
            end
        end,

        ToggleMinimap = function()
            local Tables = addon.Core.Tables;
            local minimapSettings = addonTable.minimapSettings;

            if not Tables.GetValue(minimapSettings, "hide", false) then
                minimapSettings.hide = true;

                addonIcon:Hide(dataTextName);
            else
                minimapSettings.hide = false;

                addonIcon:Show(dataTextName);

                if reloadRequired then
                    addon.API.Blizzard.StaticPopup_Show("LERH_TOGGLE_MINIMAP");
                    reloadRequired = false;
                end
            end
        end,
    };

    Tooltip = {
        Actions = function()
            local Tables = addon.Core.Tables;
            local minimapSettings = addonTable.minimapSettings;

            local changeLootSpec = Tables.GetValue(minimapSettings, "changeLootSpec", 1) == 1;
            local changeSpec = Tables.GetValue(minimapSettings, "changeSpec", 1) == 1;
            local reloadUI = Tables.GetValue(minimapSettings, "reloadUI", 1) == 1;
            local toggleIcon = Tables.GetValue(minimapSettings, "toggleIcon", 1) == 1;
            local useRotations = Tables.GetValue(minimapSettings, "useRotations", 1) == 1;
            local useTalentBuilds = addon.System.Plugins.IsActive("Talent Builds") and Tables.GetValue(minimapSettings, "useTalentBuilds", 1) == 1;

            if Mouse.Actions() then
                iconTooltip:AddLine(" ");
                iconTooltip:AddHeader(string.format("%s%s%s", HIGHLIGHT_FONT_COLOR_CODE, "Mouse Actions", addonTable.TAGS.END));
            end

            if addon.API.Blizzard.IsControlKeyDown() then
                if changeSpec then
                    iconTooltip:AddLine(string.format("%sChange Talent Specialization:%s %sControl +%s %s", addonTable.COLORS.TAGS.DEFAULT, addonTable.TAGS.END, "|cff00FF00", addonTable.TAGS.END, leftButtonTextureString));
                end

                if changeLootSpec then
                    iconTooltip:AddLine(string.format("%sChange Loot Specialization:%s %sControl +%s %s", addonTable.COLORS.TAGS.DEFAULT, addonTable.TAGS.END, "|cff00FF00", addonTable.TAGS.END, rightButtonTextureString));
                end
            elseif addon.API.Blizzard.IsShiftKeyDown() then
                if toggleIcon then
                    iconTooltip:AddLine(string.format("%sToggle Minimap Icon:%s %sShift +%s %s", addonTable.COLORS.TAGS.DEFAULT, addonTable.TAGS.END, "|cff00FF00", addonTable.TAGS.END, leftButtonTextureString));
                end

                if reloadUI then
                    iconTooltip:AddLine(string.format("%sReload UI:%s %sShift +%s %s", addonTable.COLORS.TAGS.DEFAULT, addonTable.TAGS.END, "|cff00FF00", addonTable.TAGS.END, rightButtonTextureString));
                end
            elseif not addon.API.Blizzard.IsModifierKeyDown() then
                if useRotations then
                    local apiGeneral = addon.API.General;
                    local specInfo = apiGeneral.GetSpecializationInfo();

                    if not Tables.IsEmpty(specInfo) then
                        local numRotations = addon.Rotation.GetRegisteredRotations(specInfo.ID);

                        if numRotations >= 1 then
                            iconTooltip:AddLine(string.format("%sChange Active Rotation:%s %s", addonTable.COLORS.TAGS.DEFAULT, addonTable.TAGS.END, leftButtonTextureString));
                        end
                    end
                end

                if useTalentBuilds then
                    local TalentBuilds = addon.Features.TalentBuilds;
                    local buildsAvailable = not Tables.IsEmpty(TalentBuilds.Available());
                    local changesAllowed = TalentBuilds.ChangesAllowed();

                    if buildsAvailable and changesAllowed then
                        iconTooltip:AddLine(string.format("%sChange Current Talent Build:%s %s", addonTable.COLORS.TAGS.DEFAULT, addonTable.TAGS.END, rightButtonTextureString));
                    end
                end
            end
        end,

        Builds = function()
            local Tables = addon.Core.Tables;
            local minimapSettings = addonTable.minimapSettings;

            local useTalentBuilds = addon.System.Plugins.IsActive("Talent Builds") and Tables.GetValue(minimapSettings, "useTalentBuilds", 1) == 1;

            if useTalentBuilds then
                local TalentBuilds = addon.Features.TalentBuilds;
                local buildInfo = TalentBuilds.ActiveBuild();

                local availableBuilds = TalentBuilds.Available();
                local buildsAvailable = not Tables.IsEmpty(TalentBuilds.Available());
                local changesAllowed = TalentBuilds.ChangesAllowed();

                iconTooltip:AddLine(" ");
                iconTooltip:AddLine(string.format("%s%s:%s %s%s%s", HIGHLIGHT_FONT_COLOR_CODE, "Talent Changes Allowed", addonTable.TAGS.END, (changesAllowed and addonTable.COLORS.TAGS.ACTIVE) or addonTable.COLORS.TAGS.INACTIVE, (changesAllowed and "Yes") or "No", addonTable.TAGS.END));
                iconTooltip:AddLine(string.format("%s%s:%s %s%s%s", HIGHLIGHT_FONT_COLOR_CODE, "Current Build", addonTable.TAGS.END, addonTable.COLORS.TAGS.ACTIVE, buildInfo.Name, addonTable.TAGS.END));

                iconTooltip:AddLine(" ");
                iconTooltip:AddHeader(string.format("%s%s%s", HIGHLIGHT_FONT_COLOR_CODE, "Available Talent Builds", addonTable.TAGS.END));

                if buildsAvailable then
                    for buildName in pairs(availableBuilds) do
                        iconTooltip:AddLine(string.format("%s%s%s", addonTable.COLORS.TAGS.DEFAULT, buildName, addonTable.TAGS.END));
                    end
                else
                    iconTooltip:AddLine(string.format("%s%s%s", addonTable.COLORS.TAGS.DEFAULT, "No preset builds available!", addonTable.TAGS.END));
                end
            end
        end,

        Rotation = function()
            local apiGeneral = addon.API.General;
            local apiUnit = addon.API.Unit;
            local Rotation = addon.Rotation;
            local Tables = addon.Core.Tables;

            local addonTitle, _, addonVersion = addon.GetInfo(true);

            local specInfo = apiGeneral.GetSpecializationInfo();
            local rotationInfo = Tables.GetValue(addon.currentRotation, "scriptInfo", {});

            iconTooltip:AddHeader(string.format("%s%s:%s %s%s%s", HIGHLIGHT_FONT_COLOR_CODE, addonTitle, addonTable.TAGS.END, addonTable.COLORS.TAGS.DEFAULT, addonVersion, addonTable.TAGS.END));

            iconTooltip:AddLine(" ");
            iconTooltip:AddHeader(string.format("%s%s%s", HIGHLIGHT_FONT_COLOR_CODE, "Rotation Information", addonTable.TAGS.END));

            if Tables.IsEmpty(specInfo) then
                iconTooltip:AddLine(string.format("%s%s%s", addonTable.COLORS.TAGS.DEFAULT, "No specialization selected so unable to use a rotation.", addonTable.TAGS.END));
            elseif addon.currentRotation then
                local guideAuthor = Tables.GetValue(rotationInfo, "GuideAuthor");

                if guideAuthor then
                    iconTooltip:AddLine(string.format("%sGuide Author:%s %s%s%s", addonTable.COLORS.TAGS.DEFAULT, addonTable.TAGS.END, HIGHLIGHT_FONT_COLOR_CODE, guideAuthor, addonTable.TAGS.END));
                end

                local guideLink = Tables.GetValue(rotationInfo, "GuideLink");

                if guideLink then
                    local unitClass = apiUnit.UnitClass("player");

                    if not Tables.IsEmpty(unitClass) then
                        local URL = string.format("%s|Hurl:%s|h[%s]|h%s", addonTable.COLORS.CLASS[unitClass.ID], guideLink, guideLink, addonTable.TAGS.END);

                        iconTooltip:AddLine(string.format("%sGuide Link:%s %s", addonTable.COLORS.TAGS.DEFAULT, addonTable.TAGS.END, URL));
                    end
                end

                local rotationAuthor = Tables.GetValue(rotationInfo, "ScriptAuthor");

                if rotationAuthor then
                    iconTooltip:AddLine(string.format("%sRotation Author:%s %s%s%s", addonTable.COLORS.TAGS.DEFAULT, addonTable.TAGS.END, HIGHLIGHT_FONT_COLOR_CODE, rotationAuthor, addonTable.TAGS.END));
                end

                local rotationCredits = Tables.GetValue(rotationInfo, "ScriptCredits");

                if rotationCredits then
                    iconTooltip:AddLine(string.format("%sRotation Credits:%s %s%s%s", addonTable.COLORS.TAGS.DEFAULT, addonTable.TAGS.END, HIGHLIGHT_FONT_COLOR_CODE, rotationCredits, addonTable.TAGS.END));
                end

                local rotationVersion = Tables.GetValue(rotationInfo, "WoWVersion");

                if rotationVersion then
                    iconTooltip:AddLine(string.format("%sWoW Version:%s %s", addonTable.COLORS.TAGS.DEFAULT, addonTable.TAGS.END, Tooltip.VersionNumber(rotationVersion)));
                end

                local rotationNotes = Tables.GetValue(rotationInfo, "ImportantNotes");

                if rotationNotes then
                    iconTooltip:AddLine(" ");
                    iconTooltip:AddLine(string.format("%sImportant Notes:%s %s%s%s", addonTable.COLORS.TAGS.DEFAULT, addonTable.TAGS.END, addonTable.COLORS.TAGS.DEFAULT, rotationNotes, addonTable.TAGS.END));
                end
            else
                local rotationCount = Rotation.GetRegisteredRotations(Tables.GetValue(specInfo, "ID"));

                if rotationCount >= 1 then
                    -- No Rotation is loaded.
                    iconTooltip:AddLine(string.format("%sThe addon currently has no rotation loaded, you can selected a rotation to load if you desire through the addons settings!%s", addonTable.COLORS.TAGS.DEFAULT, addonTable.TAGS.END));
                else
                    -- No Rotation exists.
                    iconTooltip:AddLine(string.format("%sThe %q rotation is not currently supported by this addon!%s", addonTable.COLORS.TAGS.DEFAULT, specInfo.Name, addonTable.TAGS.END));
                end
            end
        end,

        Setup = function()
            if iconTooltip then
                iconTooltip:Clear();

                Tooltip.Rotation();
                Tooltip.Builds();
                Tooltip.Actions();

                Tooltip.Skin(iconTooltip);

                iconTooltip:Show();
            end
        end,

        Skin = function(frame)
            local ElvUI, TukUI = ElvUI, TukUI;

            if ElvUI or TukUI then
                if frame.StripTextures then
                    frame:StripTextures();
                end

                if frame.SetTemplate then
                    frame:SetTemplate("Transparent");
                end
            end
        end,

        VersionNumber = function(versionNumber)
            local returnValue;

            if versionNumber then
                local Core = addon.Core.General;
                local majorVersion, minorVersion, revisionNumber;

                majorVersion = Core.ToNumber(string.sub(versionNumber, 1, 1));
                minorVersion = Core.ToNumber(string.sub(versionNumber, 2, 3));
                revisionNumber = Core.ToNumber(string.sub(versionNumber, 4, 5));

                if revisionNumber == 0 then
                    returnValue = string.format("%s.%s", tostring(majorVersion), tostring(minorVersion));
                else
                    returnValue = string.format("%s.%s.%s", tostring(majorVersion), tostring(minorVersion), tostring(revisionNumber));
                end

                if versionNumber < addon.CURRENT_WOW_VERSION then
                    returnValue = string.format("%s%s (Outdated)%s", addonTable.COLORS.TAGS.INACTIVE, returnValue, addonTable.TAGS.END);
                else
                    returnValue = string.format("%s%s (Current)%s", addonTable.COLORS.TAGS.ACTIVE, returnValue, addonTable.TAGS.END);
                end
            end

            return returnValue;
        end,
    };

    ---------------------- Public Functions ----------------------
    addon.Features.MinimapIcon = {
        Toggle = function()
            Mouse.ToggleMinimap();
        end,
    };

    ----------------------- Plugin Code Operation ----------------
    local moduleEvents = {
        playerLogin = function(specInfo)
            addon.System.Events.CombatEventDelay(Icon.Create, specInfo);
        end,

        changeSpecFinish = function(specInfo)
            Icon.UpdateLDB(specInfo);
        end,

        toggleMinimap = function()
            addon.Features.MinimapIcon.Toggle();
        end,

        updateMinimapIcon = function()
            local apiGeneral = addon.API.General;
            local specInfo = apiGeneral.GetSpecializationInfo();

            Icon.UpdateLDB(specInfo);
        end,
    };

    local tableEvents = {
        MODIFIER_STATE_CHANGED = Icon.UpdateToolTip,
        PLAYER_TALENT_UPDATE = Icon.UpdateToolTip,
    };

    -- These are only registered if the plugin is active as they control how what is called
    -- from LunaEclipse Rotation Helper
    addon.System.Callbacks.RegisterAddonCallbacks(moduleEvents);
    addon.System.Events.RegisterEvents(tableEvents, true);
end