-- Pulls back the Addon-Local Variables and store them locally.
local addonName, addonTable = ...;
local addon = _G["LunaEclipse_Rotation"];
local L = LibStub("AceLocale-3.0"):GetLocale(addonName);

local pluginName = "SpellFlash";

-- Check for SpellFlash and LunaEclipse Rotation Helper being loaded first so it doesn't register the plugin if its 
-- not needed.  This needs to be separate of the plugin active check as the plugin should only be registered
-- if SpellFlash is loaded, however it must be registered regardless of whether it is active or not.
if addon and addon.System.Plugins.IsActive(pluginName) then
	-- Store local copies of Blizzard API and other addon global functions and variables
	local format, huge, ipairs, pairs, print, tinsert = format, math.huge, ipairs, pairs, print, tinsert;

	local flashSpells = {};
	local priorityList = { "Interrupt", "OFFGCD", "Cooldown", "MultiTarget", "SingleTarget", "Defensive" };

	local function getFlashColor(type)
		local coreTables = addon.Core.Tables;
		local profile = coreTables.GetValue(addonTable.spellFlashSettings, "groupColors");

		return coreTables.GetValue(profile, format("color%s", type));
	end

	local function stopFlashing(flashAction)
		local coreTables = addon.Core.Tables;

		if not coreTables.IsEmpty(flashAction) then
			local actionType = coreTables.GetValue(flashAction, "Type");
			local actionID = coreTables.GetValue(flashAction, "ID");

			addon.Features.SpellFlash.Stop(actionType, actionID);
		end
	end

	local function updateFlashing(flashType, flashAction)
		local coreTables = addon.Core.Tables;

		if not coreTables.IsEmpty(flashAction) then
			local color = getFlashColor(flashType);

			local actionID = coreTables.GetValue(flashAction, "ID");
			local actionType = coreTables.GetValue(flashAction, "Type");

			addon.Features.SpellFlash.Flash(actionType, actionID, color);
		end
	end

	local function performUpdate(flashType, action)
		local coreTables = addon.Core.Tables;

		local currentAction, currentKey, currentType, iconName, newAction, previousAction, previousKey, previousType;
		local currentTypes, isFlashing, previousTypes, wasFlashing = {}, {}, {}, {};
		local previousFlashing = coreTables.Duplicate(flashSpells);

		if not coreTables.GetValue(action, "Wait", false) then
			currentAction = coreTables.GetValue(flashSpells, flashType);

			if not coreTables.IsEmpty(action) then
				newAction = {
					ID = coreTables.GetValue(action, "ID"),
					Type = coreTables.GetValue(action, "Type"),
				};
			end

			if not coreTables.Equate(currentAction, newAction) then
				flashSpells[flashType] = coreTables.Duplicate(newAction);
			end

			for _, key in ipairs(priorityList) do
				currentAction = flashSpells[key];
				previousAction = previousFlashing[key];

				currentKey = format("%s-%s", coreTables.GetValue(currentAction, "Type", ""), coreTables.GetValue(currentAction, "ID", ""));
				previousKey = format("%s-%s", coreTables.GetValue(previousAction, "Type", ""), coreTables.GetValue(previousAction, "ID", ""));

				if not wasFlashing[previousKey] then
					wasFlashing[previousKey] = previousAction;
					previousTypes[previousKey] = key;
				end

				if not isFlashing[currentKey] then
					isFlashing[currentKey] = currentAction;
					currentTypes[currentKey] = key;
				end
			end

			for previousKey, flashAction in pairs(wasFlashing) do
				currentType = currentTypes[previousKey] or huge;
				previousType = previousTypes[previousKey];

				if not isFlashing[previousKey] or coreTables.GetKey(priorityList, previousType) < coreTables.GetKey(priorityList, currentType) then
					stopFlashing(flashAction);

					wasFlashing[previousKey] = nil;
				end
			end

			for currentKey, flashAction in pairs(isFlashing) do
				currentType = currentTypes[currentKey];

				if not wasFlashing[currentKey] then
					updateFlashing(currentType, flashAction);
				end
			end
		end
	end

	local function leaveCombat()
		for _, flashAction in pairs(flashSpells) do
			stopFlashing(flashAction);
		end
	end

	local moduleEvents = {
		Disabled = function()
			-- Any code to run when there is no longer a rotation active.
			print(L["No rotation available, deactivating SpellFlashing"]);
			leaveCombat();
		end,

		Enabled = function()
			print(L["Rotation available, activating SpellFlashing"]);
			leaveCombat();
		end,

		CooldownIcon = function(ActionList)
			-- This only happens when the Cooldown Icons action list has changed.
			-- Sometimes the first spell will still be the first spell, this just means the entire list changed somehow.
			-- ActionList is the queue for the icon, consider passing it to your CreateFlashTableByFlashType
			performUpdate("Cooldown", addon.Core.Tables.GetValue(ActionList, 1));
		end,

		DefensiveIcon = function(ActionList)
			-- This only happens when the Defensive Icons action list has changed.
			-- Sometimes the first spell will still be the first spell, this just means the entire list changed somehow.
			-- ActionList is the queue for the icon, consider passing it to your CreateFlashTableByFlashType
			performUpdate("Defensive", addon.Core.Tables.GetValue(ActionList, 1));
		end,

		InterruptIcon = function(ActionList)
			-- This only happens when the Interrupt Icons action list has changed.
			-- Sometimes the first spell will still be the first spell, this just means the entire list changed somehow.
			-- ActionList is the queue for the icon, consider passing it to your CreateFlashTableByFlashType
			performUpdate("Interrupt", addon.Core.Tables.GetValue(ActionList, 1));
		end,

		OffGCDIcon = function(ActionList)
			-- This only happens when the OffGCD Icons action list has changed.
			-- Sometimes the first spell will still be the first spell, this just means the entire list changed somehow.
			-- ActionList is the queue for the icon, consider passing it to your CreateFlashTableByFlashType
			performUpdate("OFFGCD", addon.Core.Tables.GetValue(ActionList, 1));
		end,

		MultiTargetIcon = function(ActionList)
			-- This only happens when the Multi Target Icons action list has changed.
			-- Sometimes the first spell will still be the first spell, this just means the entire list changed somehow.
			-- ActionList is the queue for the icon, consider passing it to your CreateFlashTableByFlashType
			performUpdate("MultiTarget", addon.Core.Tables.GetValue(ActionList, 1));
		end,

		SingleTargetIcon = function(ActionList)
			-- This only happens when the Single Target Icons action list has changed.
			-- Sometimes the first spell will still be the first spell, this just means the entire list changed somehow.
			-- ActionList is the queue for the icon, consider passing it to your CreateFlashTableByFlashType
			performUpdate("SingleTarget", addon.Core.Tables.GetValue(ActionList, 1));
		end,
	};

	local tableEvents = {
		PLAYER_REGEN_ENABLED = leaveCombat,
	};

	-- These are only registered if the plugin is active as they control how what is called
	-- from LunaEclipse Rotation Helper
	addon.System.Callbacks.RegisterAddonCallbacks(moduleEvents);
	addon.System.Events.RegisterEvents(tableEvents);
end